using System;
using System.Collections.Generic;
using System.Security.Cryptography;
using System.Text;
using eap;
using gsm.datasource;
using log4net;
using Utility;

namespace vpnsim.authenticator.core
{
    /// <summary>
    /// A class for performing some basic sim functions required by the authentication server
    /// </summary>
    public class SimFunctions : ISimFunctions
    {
        private AuthenticationServer authServer;
        private static readonly ILog log = LogManager.GetLogger(typeof(SimFunctions));

        /// <summary>
        /// Constructor for the Simfunction class
        /// </summary>
        /// <param name="authServer">A reference to the authentication server this class is associated to</param>
        public SimFunctions(AuthenticationServer authServer)
        {
            this.authServer = authServer;
        }

        /// <summary>
        /// Register a key in the authentication server database
        /// </summary>
        /// <param name="imsi">The username to register</param>
        /// <param name="key">The key to register</param>
        public void RegisterKey(byte[] imsi, byte[] key)
        {
            authServer.AuthenticateUser(HexEncoding.IMSIConvert(imsi), key);
        }

        /// <summary>
        /// Not yet implemented. Validates the IMSI according to basic IMSI structure guidelines. Country code .. and so on. Valid subscriber?
        /// </summary>
        /// <param name="imsi">The IMSI to validate</param>
        /// <returns>Whether the IMSI was valid or not</returns>
        public bool ValidateIMSI(byte[] imsi)
        {
            //TODO Fixme
            TripletSource ts = new TripletSource();
            return ts.ValidateIMSI(imsi);
        }

        /// <summary>
        /// A function that calculates the SRES to a given RAND. Not used here.
        /// </summary>
        /// <param name="rands">Array of rands to calculate SRES for</param>
        /// <param name="sres">Return byte array of sres results</param>
        /// <param name="Kc">The key calculated</param>
        /// <returns>Whether the rands were valid</returns>
        public bool CalculateSRES(byte[] rands, out byte[] sres, out byte[] Kc)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Look up the rands for a given IMSI. Should query the Identity Provider and fetch Authentication Triplets. This function looks up static values.
        /// </summary>
        /// <param name="imsi">The IMSI to return rands for</param>
        /// <param name="number">The number of rands to return</param>
        /// <param name="sres">Return byte array of sres values</param>
        /// <param name="Kc">Return byte array of Kc values</param>
        /// <returns>Return a byte array of rand values</returns>
        public byte[] GetRANDs(byte[] imsi, int number, out byte[] sres, out byte[] Kc)
        {
            TripletSource ts = new TripletSource();
            byte[] returnRand = new byte[16*number]; //Each rand is 16 bytes
            byte[] returnSres = new byte[4*number]; //Each sres is 4 bytes 
            byte[] returnKc = new byte[8*number]; //Each kc is 8 bytes
            
            for(int i=0;i<number;i++)
            {
                byte[] tempRand, tempSres, tempKc;
                ts.ReadTriplet(imsi, out tempRand, out tempSres, out tempKc);
                Array.Copy(tempRand, 0, returnRand, 16*i, 16);
                Array.Copy(tempSres, 0, returnSres, 4*i, 4);
                Array.Copy(tempKc, 0, returnKc, 8*i, 8);
            }

            sres = returnSres;
            Kc = returnKc;
            return returnRand;
        }

        /// <summary>
        /// Calculate the master key using SHA1 based on the values given
        /// </summary>
        /// <param name="imsi">The IMSI of the subscriber</param>
        /// <param name="kc">The Kc values of the auth. vectors</param>
        /// <param name="nonce">Nonce value sent by subscriber</param>
        /// <param name="verList">List of supported versions given</param>
        /// <param name="selectedVersion">The selected version</param>
        /// <returns>Returns a byte array which is the master key</returns>
        public byte[] GetMasterKey(byte[] imsi, byte[] kc, byte[] nonce, short[] verList, short selectedVersion)
        {
            byte[] data = new byte[imsi.Length + kc.Length + nonce.Length + verList.Length * 2 + 2];
            Array.Copy(imsi, 0, data, 0, imsi.Length);
            Array.Copy(kc, 0, data, imsi.Length, kc.Length);
            Array.Copy(nonce, 0, data, imsi.Length + kc.Length, nonce.Length);
            byte[] verListBytes = new byte[verList.Length * 2];
            for (int i = 0; i < verList.Length; i++)
            {
                verListBytes[2 * i] = (byte)((verList[i] & 0xFF00) >> 8);
                verListBytes[2 * i + 1] = (byte)((verList[i] & 0x00FF));
            }
            byte[] selVerBytes = new byte[2];
            selVerBytes[0] = (byte)((selectedVersion & 0xFF00) >> 8);
            selVerBytes[1] = (byte)((selectedVersion & 0x00FF));
            Array.Copy(verListBytes, 0, data, imsi.Length + kc.Length + nonce.Length, verListBytes.Length);
            Array.Copy(selVerBytes, 0, data, imsi.Length + kc.Length + nonce.Length + verListBytes.Length, 2);
            SHA1 sha = new SHA1CryptoServiceProvider();
            byte[] result = sha.ComputeHash(data);
            return result;
        }
    }
}
