using System;
using System.Collections.Generic;
using System.Text;

using eap;
using log4net;
using Org.Mentalis.Security.Cryptography;

namespace eap
{
    /// <summary>
    /// Class for handling incoming EAP messages
    /// </summary>
    public class EAPHandler
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(EAPHandler));
        private ISimFunctions simAuth;
        private Random identifier;
        public byte[] imsi;
        public byte[] nonce;
        private byte[] sres;
        private byte[] Kc;
        private byte[] dhKey;
        public short selectedVersion;
        public short[] versionList;
        private byte[] masterKey;
        private DiffieHellman dh;

        /// <summary>
        /// Constructor for the EAPHandler class
        /// </summary>
        /// <param name="simAuth">A class implementing the ISimFunctions interface</param>
        public EAPHandler(ISimFunctions simAuth)
        {
            identifier = new Random();
            this.simAuth = simAuth;
        }
        private byte GetIdentifier()
        {
            return (byte) identifier.Next();
        }

        /// <summary>
        /// The EAP Message parser class. Takes a standard EAPMessage, and determines whether it is a subtype. Then generates the correct response to the message.
        /// </summary>
        /// <param name="message">The EAPMessage do parse</param>
        /// <returns>The message response</returns>
        public EAPMessage HandleMessage(EAPMessage message)
        {
            EAPMessage response = new EAPMessage(Code.Failure);
            if (message.code.Equals(Code.Response)) //Is it an EAP-Request?
            {
                // Pass it on to the Responsehandler
                response = HandleResponse((EAPRequestResponseMessage)message);
            }else if(message.code.Equals(Code.Request)) // Or an EAP-Response?
            {
                // Pass it on to the Requesthandler
                response = HandleRequest((EAPRequestResponseMessage) message);
            }else if(message.code.Equals(Code.Success)) // Or EAP-Success?
            {
                // Respond with a Successmessage
                response = new EAPSuccessFailureMessage(Code.Success);
            }
            return response;

        }

        /// <summary>
        /// Handles EAP Responses
        /// </summary>
        /// <param name="message">The EAPRequestResponseMessage to parse</param>
        /// <returns>The message to return</returns>
        private EAPMessage HandleResponse(EAPRequestResponseMessage message)
        {
            EAPMessage response = new EAPSuccessFailureMessage(Code.Failure);

            // If the EAP-type is SIM, cast and pass it on to the SIMhandler.
            if(message.eapType.Equals(EAPType.SIM))
            {
                EAPSIMMessage inMsg = (EAPSIMMessage)message;
                response = HandleSIMResponse(inMsg);
            }
            return response;
        }

        /// <summary>
        /// Handles EAP-Response/SIM messages
        /// </summary>
        /// <param name="message">The EAPSIMMessage to parse</param>
        /// <returns>The message to return</returns>
        private EAPMessage HandleSIMResponse(EAPSIMMessage message)
        {
            EAPMessage response = new EAPSuccessFailureMessage(Code.Failure);

            // If it is a EAP-Response/SIM/Start-message, parse accordingly
            if (message.subtype.Equals(EAPSIMMessage.Subtype.Start))
            {
                EAPResponseStart inMsg = (EAPResponseStart)message;
                //Get the attributes and store them for later use
                imsi = inMsg.GetIdentity();
                nonce = inMsg.GetNonce();
                selectedVersion = inMsg.GetSelectedVersion();
                versionList = new short[] {selectedVersion};

                //If the IMSI is not valid (for this authenticator) return a failure message
                if(!simAuth.ValidateIMSI(imsi))
                {
                    log.Info("IMSI not recognized/denied by tripletserver"+BitConverter.ToString(imsi));
                    EAPSuccessFailureMessage fail = new EAPSuccessFailureMessage(Code.Failure);
                    response = fail;
                }else
                {
                    EAPRequestChallenge outMsg = new EAPRequestChallenge();
                    outMsg.SetRAND(simAuth.GetRANDs(imsi, 3, out sres, out Kc));

                    masterKey = simAuth.GetMasterKey(imsi, Kc, nonce, versionList, selectedVersion);
                    outMsg.SetIdentifier((byte)identifier.Next());
                    outMsg.GenerateMAC(nonce, masterKey);
                    response = outMsg;
                }
                // If it is a EAP-Response/SIM/Challenge-message, parse accordingly
            }else if(message.subtype.Equals(EAPSIMMessage.Subtype.Challenge))
            {
                EAPResponseChallenge inMsg = (EAPResponseChallenge) message;
                byte[] MAC = inMsg.GetMAC();
                if(!inMsg.VerifyMAC(nonce, masterKey)) //Mac not valid!
                {
                    log.Debug("Invalid mac received");
                    EAPSuccessFailureMessage outMsg = new EAPSuccessFailureMessage(Code.Failure);
                    outMsg.SetIdentifier(inMsg.GetIdentifier());
                    response = outMsg;
                }else
                {
                    EAPRequestKeyExchange outMsg = new EAPRequestKeyExchange();
                    dh = new DiffieHellmanManaged();
                    DHParameters dhParams = dh.ExportParameters(false);
                    outMsg.SetKeyattributes(dh.CreateKeyExchange(), dhParams);
                    outMsg.SetIdentifier((byte)identifier.Next());
                    response = outMsg;
                }

                // If it is a EAP-Response/SIM/KeyExchange-message, parse accordingly
            }else if(message.subtype.Equals(EAPSIMMessage.Subtype.KeyExchange))
            {
                EAPResponseKeyExchange inMsg = (EAPResponseKeyExchange) message;

                //Now we can find the key
                dhKey = dh.DecryptKeyExchange(inMsg.GetKey());
                simAuth.RegisterKey(imsi, dhKey);

                EAPSuccessFailureMessage outMsg = new EAPSuccessFailureMessage(Code.Success);
                response = outMsg;
            }
            
            return response;
        }

        /// <summary>
        /// Handles EAP Requests
        /// </summary>
        /// <param name="message">The EAPRequestResponseMessage to parse</param>
        /// <returns>The message to return</returns>
        private EAPMessage HandleRequest(EAPRequestResponseMessage message)
        {
            EAPMessage response = new EAPSuccessFailureMessage(Code.Failure);
            if( message.eapType.Equals(EAPType.SIM))
            {
                EAPSIMMessage inMsg = (EAPSIMMessage)message;
                response = HandleSIMRequest(inMsg);
            }else if(message.eapType.Equals(EAPType.KeyExchange))
            {
                EAPRequestKeyExchange inMsg = (EAPRequestKeyExchange) message;
                EAPResponseKeyExchange outMsg = new EAPResponseKeyExchange();
                outMsg.SetIdentifier(inMsg.GetIdentifier());
                response = outMsg;
            }
            return response;
        }

        /// <summary>
        /// Handles EAP-Request/SIM messages
        /// </summary>
        /// <param name="message">The EAPSIMMessage to parse</param>
        /// <returns>The message to return</returns>
        private EAPMessage HandleSIMRequest(EAPSIMMessage message)
        {
            EAPMessage response = new EAPSuccessFailureMessage(Code.Failure);
            //Is it of type EAP-Request/SIM/Start? parse accordingly
            if (message.subtype.Equals(EAPSIMMessage.Subtype.Start))
            {
                EAPRequestStart inMsg = (EAPRequestStart) message;
                EAPResponseStart outMsg = new EAPResponseStart();
                outMsg.SetIdentifier(inMsg.GetIdentifier());
                response = outMsg;

                //Is it of type EAP-Request/SIM/Challenge? parse accordingly
            }else if(message.subtype.Equals(EAPSIMMessage.Subtype.Challenge))
            {
                EAPRequestChallenge inMsg = (EAPRequestChallenge) message;
                byte[] rands = inMsg.GetRand();
                bool res = simAuth.CalculateSRES(rands, out sres, out Kc);

                masterKey = simAuth.GetMasterKey(imsi, Kc, nonce, versionList, selectedVersion);
                
                if(!inMsg.VerifyMAC(nonce, masterKey)) //Mac not valid!
                {
                    log.Debug("Invalid MAC received");
                    EAPSuccessFailureMessage outMsg = new EAPSuccessFailureMessage(Code.Failure);
                    outMsg.SetIdentifier(inMsg.GetIdentifier());
                    response = outMsg;
                }else
                {
                    EAPResponseChallenge outMsg = new EAPResponseChallenge();
                    outMsg.SetIdentifier(inMsg.GetIdentifier());
                    outMsg.GenerateMAC(nonce, masterKey);
                    response = outMsg;
                }

                //Is it of type EAP-Request/SIM/KeyExchange? parse accordingly
            }else if(message.subtype.Equals(EAPSIMMessage.Subtype.KeyExchange))
            {
                EAPRequestKeyExchange inMsg = (EAPRequestKeyExchange) message;
                DHParameters dhp = inMsg.GetKeyParameters();
                dh = new DiffieHellmanManaged(dhp.P,dhp.G,160);
                
                EAPResponseKeyExchange outMsg = new EAPResponseKeyExchange();
                outMsg.SetKey(dh.CreateKeyExchange());
                outMsg.SetIdentifier(inMsg.GetIdentifier());
                
                //Now we can find the key
                dhKey = dh.DecryptKeyExchange(inMsg.GetKey());
                
                simAuth.RegisterKey(imsi, dhKey);

                response = outMsg;
            }
            return response;
        }

    }
}
