using System;
using System.Collections.Generic;
using System.Security.Cryptography;
using System.Text;
using log4net;

namespace eap
{
    /// <summary>
    /// Enum for the message code
    /// </summary>
    public enum Code : byte
    {
        Request = 1,
        Response = 2,
        Success = 3,
        Failure = 4
    }

    /// <summary>
    /// Enum for the EAP Type
    /// </summary>
    public enum EAPType : byte
    {
        Identity = 1,
        SIM = 18,
        KeyExchange = 255
    }        

    /// <summary>
    /// Basic class for representing an EAP message
    /// </summary>
    public class EAPMessage
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(EAPMessage));
        public Code code;
        private byte identifier = 0;
        public byte[] data = new byte[0];
        public EAPMessage(Code code)
        {
            this.code = code;
        }

        /// <summary>
        /// Returns the actual length of the eap message
        /// </summary>
        /// <returns></returns>
        public int Length()
        {
            return 4 + data.Length;
        }

        /// <summary>
        /// Returns the identifier of the message
        /// </summary>
        /// <returns></returns>
        public byte GetIdentifier()
        {
            return identifier;
        }

        /// <summary>
        /// Set the identifier of the message
        /// </summary>
        /// <param name="id"></param>
        public void SetIdentifier(byte id)
        {
            this.identifier = id;
        }

        /// <summary>
        /// Get the actual bytes of the eap message
        /// </summary>
        /// <returns></returns>
        public byte[] GetBytes()
        {
            byte[] bytes = new byte[Length()];
            bytes[0] = (byte)code;
            bytes[1] = identifier;
            byte[] length = BitConverter.GetBytes((char)Length());
            Array.Reverse(length);
            Array.Copy(length, 0, bytes, 2, 2);
            Array.Copy(data,0,bytes,4,data.Length);
            return bytes;
        }

        /// <summary>
        /// Print the eap-message in a pretty way (e.g EAP-Request (Attributes))
        /// </summary>
        /// <returns></returns>
        new public string ToString()
        {
            string result = "EAP-"+this.code;
            if (code.Equals(Code.Request) || code.Equals(Code.Response))
            {
                EAPRequestResponseMessage tempRR = (EAPRequestResponseMessage)this;
                result = tempRR.ToString();
                if (tempRR.eapType.Equals(EAPType.SIM)) result = ((EAPSIMMessage)tempRR).ToString();
            }
            result +=", Identifier: "+this.GetIdentifier();
            return result;
        }

        /// <summary>
        /// Parse a byte sequence in order to recreate the eap message
        /// </summary>
        /// <param name="bytes"></param>
        /// <returns></returns>
        public static EAPMessage ParseBytes(byte[] bytes)
        {
            Code code = (Code)bytes[0];
            EAPMessage message = new EAPMessage(code);
            Array.Resize(ref message.data, bytes.Length - 4);
            Array.Copy(bytes,4,message.data,0,message.data.Length);
            if(code.Equals(Code.Response) || code.Equals(Code.Request))
            {
                message = EAPRequestResponseMessage.ParseMessage(message);
            }else if(code.Equals(Code.Success) || code.Equals(Code.Failure))
            {
                message = new EAPSuccessFailureMessage(code);
            }
            message.identifier = bytes[1];
            return message;
        }
    }

    /// <summary>
    /// Class representing an EAP-Success or EAP-Failure message
    /// </summary>
    public class EAPSuccessFailureMessage : EAPMessage
    {
        public EAPSuccessFailureMessage(Code code) : base(code){}
    }

    /// <summary>
    /// Class representing an EAP-Request or EAP-Response message
    /// </summary>
    public class EAPRequestResponseMessage : EAPMessage
    {
        public EAPType eapType
        {
            get { return (EAPType)data[0]; }
            set { data[0] = (byte)value; }
        }
        new public string ToString()
        {
            return "EAP-" + code + "/" + eapType;
        }

        public EAPRequestResponseMessage(Code code, EAPType eapType) : base(code)
        {
            Array.Resize(ref data,1);
            this.eapType = eapType;
        }
        
        /// <summary>
        /// Parse an EAP-Request/Response-message
        /// </summary>
        /// <param name="message"></param>
        /// <returns></returns>
        public static EAPRequestResponseMessage ParseMessage(EAPMessage message)
        {
            byte[] data = message.data;
            EAPType eapType = (EAPType) data[0];
            EAPRequestResponseMessage newMessage = new EAPRequestResponseMessage(message.code, eapType);
            Array.Resize(ref newMessage.data, data.Length);
            Array.Copy(data, newMessage.data, data.Length);
            
            if (newMessage.eapType.Equals(EAPType.SIM)) newMessage = EAPSIMMessage.ParseMessage(newMessage);
            return newMessage;
        }
    }
    }
