using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace eap
{
    /// <summary>
    /// Class representing an attribute of a EAP-SIM message
    /// </summary>
    public class EAPSIMAttribute
    {
        /// <summary>
        /// An enum representing the attribute types available
        /// </summary>
        public enum AttributeType : byte
        {
            AT_NOT_SET = 0,
            AT_VERSION_LIST = 15,
            AT_SELECTED_VERSION = 16,
            AT_NONCE_MT = 7,
            AT_PERMANENT_ID_REQ = 10,
            AT_ANY_ID_REQ = 13,
            AT_FULLAUTH_ID_REQ = 17,
            AT_IDENTITY = 14,
            AT_RAND = 1,
            AT_NEXT_PSEUDONYM = 132,
            AT_NEXT_REAUTH_ID = 133,
            AT_IV = 129,
            AT_ENCR_DATA = 130,
            AT_PADDING = 6,
            AT_RESULT_IND = 135,
            AT_MAC = 11,
            AT_COUNTER = 19,
            AT_COUNTER_TOO_SMALL = 20,
            AT_NONCE_S = 21,
            AT_NOTIFICATION = 12,
            AT_CLIENT_ERROR_CODE = 22,
            AT_KEYPARAMG = 253,
            AT_KEYPARAMP = 254,
            AT_PUBLICKEY = 255
            
        }
        private byte[] tlvData;
        private AttributeType type
        {
            set { tlvData[0] = (byte) value; }
            get { return (AttributeType)tlvData[0]; }
        }
        private byte length
        {
            set { tlvData[1] = value; }
            get { return tlvData[1]; }
        }
        private byte[] value
        {
            get 
            { 
                byte[] val = new byte[tlvData.Length-2];
                Array.Copy(tlvData, 2,val,0,val.Length);
                return val;
            }
            set
            {
                Array.Copy(value,0,tlvData,2,value.Length);
            }
            
        }
        
        /// <summary>
        /// Constructor for the attribute
        /// </summary>
        /// <param name="type">The AttributeType</param>
        /// <param name="value">The value of the attribute</param>
        public EAPSIMAttribute(AttributeType type, byte[] value)
        {
            int tlvLength = 2 + value.Length;
            //Make tlvdata size a multiple of 4 bytes
            //new byte[8 + apduBytes.Length + (4-(8+apduBytes.Length)%4)%4];

            tlvData = new byte[(((tlvLength - 1) / 4) + 1) * 4];
            this.type = type;
            this.value = value;
            this.length = (byte)(tlvData.Length/4);
        }

        /// <summary>
        /// Overloaded constructor for the attribute, takes a parameter specifying what to do with the reserved-field
        /// </summary>
        /// <param name="type">The attributetype</param>
        /// <param name="data">Data of the attribute</param>
        /// <param name="insertLength">Bool specifying whether to include actual data length or not</param>
        public EAPSIMAttribute(AttributeType type, byte[] data, bool insertLength)
        {
            short length = (short)data.Length;
            byte[] actualLength = BitConverter.GetBytes(length);
            byte[] value = new byte[2 + data.Length];
            if(insertLength) Array.Copy(actualLength, 0, value, 0, 2); //Else they're reserved
            Array.Copy(data, 0, value, 2, data.Length);

            int tlvLength = 2 + value.Length;
            //Make tlvdata size a multiple of 4 bytes
            tlvData = new byte[(((tlvLength - 1) / 4) + 1) * 4];
            this.type = type;
            this.length = (byte)(tlvData.Length/4);
            this.value = value;
            
        }

        /// <summary>
        /// GEt the size of the attribute (in multiples of 4)
        /// </summary>
        /// <returns></returns>
        public int Length() { return length; }
        
        /// <summary>
        /// Get the byte-representation of the attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetBytes()
        {
            //length = (byte)length();
            return tlvData;
        }

        /// <summary>
        /// Parse a byte array and create the appropriate attributes of the 
        /// </summary>
        /// <param name="typeData">The typedata byte array to parse</param>
        /// <returns>An array of attributes found in the byte array</returns>
        public static EAPSIMAttribute[] ParseAttributes(byte[] typeData)
        {
            ArrayList attributes = new ArrayList();
            AttributeType type;
            int length;
            byte[] value;
            for(int i=0;i<typeData.Length;)
            {
                type = (AttributeType)typeData[i+0];
                length = typeData[(i+1)]*4; // 4 bytes pr "length unit"
                value = new byte[length-2];
                Array.Copy(typeData, i + 2, value, 0, value.Length);
                attributes.Add(new EAPSIMAttribute(type, value));
                i += length;
            }
            EAPSIMAttribute[] attrs = new EAPSIMAttribute[attributes.Count];
            for(int i = 0;i<attrs.Length;i++)
            {
                attrs[i] = (EAPSIMAttribute)attributes[i];
            }
            return attrs;
        }

        /// <summary>
        /// Return the data of the attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetData()
        {
            return value;
        }

        /// <summary>
        /// Get the actual attribute value, trimmed to the length specified by the attribute
        /// </summary>
        /// <param name="withLength"></param>
        /// <returns></returns>
        public byte[] GetValue(bool withLength) //Gets the actual value of the TLV data (based on count or not)
        {
            byte[] ret;
            if (withLength)
            {
                byte[] actualLength = new byte[2];
                Array.Copy(value, 0, actualLength, 0, 2);
                ret = new byte[BitConverter.ToUInt16(actualLength, 0)];
                Array.Copy(value, 2, ret, 0, ret.Length);
            }else
            {
                ret = new byte[value.Length-2];
                Array.Copy(value,2,ret,0,ret.Length);
            }
            return ret;
        }

        /// <summary>
        /// Get the type of the attribute
        /// </summary>
        /// <returns></returns>
        public AttributeType GetAttributeType()
        {
            return this.type;
        }
    }
}
