using System;
using System.Collections.Generic;
using System.Security.Cryptography;
using System.Text;
using log4net;
using Org.Mentalis.Security.Cryptography;

namespace eap
{
    /// <summary>
    /// A class describing the EAP-../SIM message, builds on http://www.ietf.org/rfc/rfc4186.txt 
    /// and http://www.ietf.org/rfc/rfc3748.txt
    /// </summary>
    public class EAPSIMMessage : EAPRequestResponseMessage
    {
        /// <summary>
        /// A enum describing the Subtype attribute
        /// </summary>
        public enum Subtype : byte
        {
            Start = 10,
            Challenge = 11,
            Notification = 12,
            Reauthentication = 13,
            ClientError = 14,
            KeyExchange = 255
        }

        /// <summary>
        /// The subtype parameter, an alias for data[1]
        /// </summary>
        public Subtype subtype
        {
            get { return (Subtype)data[1]; }
            set { data[1] = (byte)value; }
        }

        public EAPSIMAttribute[] attributes;
        
        /// <summary>
        /// The typedata parameter, an alias for the last bytes of data
        /// </summary>
        protected byte[] typeData
        {
            get 
            { 
                byte[] td = new byte[data.Length-4];
                Array.Copy(data,4,td,0,td.Length);
                return td;
            }
            set
            {
                Array.Resize(ref data, 4+value.Length);
                Array.Copy(value,0,data,4,value.Length);
            }
        }

        /// <summary>
        /// Constructor for EAPSIMMessage
        /// </summary>
        /// <param name="code">The message code to use</param>
        /// <param name="eapType">The EAP-Type to use</param>
        /// <param name="subtype">The Subtype parameter to use</param>
        protected EAPSIMMessage(Code code, EAPType eapType, Subtype subtype) : base(code, eapType)
        {
            this.attributes = new EAPSIMAttribute[0];
            Array.Resize(ref data, 4);
            this.subtype = subtype;
        }
        /// <summary>
        /// Update the value of an existing attribute
        /// </summary>
        /// <param name="updatedAttribute">The attribute to update, with the new value set</param>
        /// <returns>Whether the attribute existed, and was found</returns>
        protected bool UpdateAttribute(EAPSIMAttribute updatedAttribute)
        {
          
            bool attributeFound = false;
            for (int i = 0; i < attributes.Length; i++)
            {
                EAPSIMAttribute attribute = attributes[i];
                bool equalAttributes = attribute.GetAttributeType().Equals(attribute.GetAttributeType());
                bool equalSize = attribute.Length().Equals(updatedAttribute.Length());
                if (equalAttributes && equalSize)
                {
                    attributes[i] = updatedAttribute;
                    attributeFound = true;
                    break;
                }
            }
            if(!attributeFound) return false;

            //We have the attribute, now update typedata
    
            //Attributes need to be of same size
            byte[] td = typeData;
            int startIndex = 0;
            attributeFound = false;
            while(startIndex<td.Length)
            {
                EAPSIMAttribute.AttributeType type = (EAPSIMAttribute.AttributeType)td[startIndex];
                //Jump out when we've found the attribute startindex
                if (type.Equals(updatedAttribute.GetAttributeType()))
                {
                    attributeFound = true;
                    break;
                }
                int length = td[startIndex + 1]; 
                startIndex += length*4; //Length is in blocks of 4 bytes
            }

            if(attributeFound)
            {
                Array.Copy(updatedAttribute.GetBytes(),0,td,startIndex,updatedAttribute.GetBytes().Length);
                typeData = td;
                return true;
            }else
            {
                //Should never be here
                return false;
            }
        }
        /// <summary>
        /// Add a new attribute to the EAP/SIM message. Do not use this to modify existing attributes, use UpdateAttribute()
        /// </summary>
        /// <param name="attribute">The new attribute to add</param>
        protected void AddAttribute(EAPSIMAttribute attribute)
        {
            Array.Resize(ref attributes, attributes.Length + 1);
            attributes[attributes.Length - 1] = attribute;
            byte[] td = typeData;
            int attributeByteLength = attribute.Length()*4;
            Array.Resize(ref td, td.Length + attributeByteLength);
            Array.Copy(attribute.GetBytes(), 0, td, td.Length - attributeByteLength, attributeByteLength);
            this.typeData = td;

        }

        /// <summary>
        /// Static function for parsing of EAP/SIM messages. Used by EAPHandler.
        /// </summary>
        /// <param name="message">The EAPRequestResponseMessage to parse</param>
        /// <returns>An EAPSIMMessage cast to the correct type, depending on the subtype</returns>
        public static EAPSIMMessage ParseMessage(EAPRequestResponseMessage message)
        {
            Subtype subType = (Subtype)message.data[1];
            EAPSIMMessage newMessage = new EAPSIMMessage(message.code, message.eapType, subType);

            if(message.code.Equals(Code.Request))
            {
                if (subType.Equals(Subtype.Start)) newMessage =  new EAPRequestStart();
                else if (subType.Equals(Subtype.Challenge)) newMessage = new EAPRequestChallenge();
                else if (subType.Equals(Subtype.Notification)) newMessage = new EAPRequestNotification();
                else if (subType.Equals(Subtype.KeyExchange)) newMessage = new EAPRequestKeyExchange();
            }else if(message.code.Equals(Code.Response))
            {
                if (subType.Equals(Subtype.Start)) newMessage = new EAPResponseStart();
                else if (subType.Equals(Subtype.Challenge)) newMessage = new EAPResponseChallenge();
                else if (subType.Equals(Subtype.Notification)) newMessage = new EAPResponseNotification();
                else if (subType.Equals(Subtype.KeyExchange)) newMessage = new EAPResponseKeyExchange();
            }
            byte[] data = message.data;
            Array.Resize(ref newMessage.data, data.Length);
            Array.Copy(data, newMessage.data, newMessage.data.Length);
            newMessage.attributes = EAPSIMAttribute.ParseAttributes(newMessage.typeData);
            return newMessage;
        }

        /// <summary>
        /// A function for calculating the MAC_RAND value
        /// </summary>
        /// <param name="key">The key for the HMACMD5-function</param>
        /// <param name="rand">The RAND bytes</param>
        /// <param name="nonce">The Nonce bytes</param>
        /// <param name="lifetime">The key lifetime</param>
        private void CalculateMAC_RAND(byte[] key, byte[] rand, byte[] nonce, byte[] lifetime)
        {
            HMACMD5 hmac = new HMACMD5(key);
            byte[] msg = new byte[rand.Length + 20];
            Array.Copy(rand, msg, rand.Length);
            Array.Copy(nonce, 0, msg, rand.Length, 16);
            Array.Copy(lifetime, 0, msg, msg.Length - 4, 4);
            byte[] hash = hmac.ComputeHash(msg);
            Array.Copy(hash, 0, data, 4, hash.Length);
        }

        /// <summary>
        /// Function for printing the EAP-SIM messages in a pretty way (eg. EAP-Request/SIM/Start (Attribute)
        /// </summary>
        /// <returns>The string to print</returns>
        public new string ToString()
        {
            string attrs = "";
            if( attributes.Length!=0)
            {
                attrs = "(";
                attrs += attributes[0].GetAttributeType();
                for (int i = 1; i < attributes.Length;i++ )
                {
                    attrs += ", ";
                    attrs += attributes[i].GetAttributeType();        
                }
                attrs += ")";
            }
            return "EAP-" + code + "/" + eapType+"/"+subtype + attrs;
        }
        /// <summary>
        /// Message for retrieving a given attribute from the SIM-message
        /// </summary>
        /// <param name="type">The EAPSIMAttribute.Attribute-type to retrieve</param>
        /// <returns>The attribute, if found. If not, a AT_NOT_SET-attribute is returned.</returns>
        public EAPSIMAttribute GetAttribute(EAPSIMAttribute.AttributeType type)
        {
            foreach(EAPSIMAttribute attribute in attributes)
            {
                if (attribute.GetAttributeType().Equals(type))
                {
                    return attribute;
                }
            }
            return new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_NOT_SET, new byte[1024],false);
        }
    }

    /// <summary>
    /// A class that represents the EAP-Request/SIM/Start message
    /// </summary>
    public class EAPRequestStart : EAPSIMMessage        // Create an EAP-Request/SIM/Start message
    {        
        public EAPRequestStart() : base(Code.Request, EAPType.SIM, Subtype.Start)
        {
        }

        /// <summary>
        /// Set the AT_VERSION_LIST attribute
        /// </summary>
        /// <param name="verlist">A list of the supported versions</param>
        public void SetVersions(short[] verlist) 
        {
            byte[] versionList = new byte[verlist.Length*2];
            byte[] version = new byte[2];
            for (int i = 0; i < verlist.Length; i = i + 2)
            {
                version = BitConverter.GetBytes((short)verlist[i]);
                Array.Copy(version, 0, versionList, 2 + i * 2, 2);
            } 
            EAPSIMAttribute versionListAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_VERSION_LIST, versionList, true);
            AddAttribute(versionListAttribute);
        }

        /// <summary>
        /// Get the AT_VERSION_LIST attribute
        /// </summary>
        /// <returns>The versions supported</returns>
        public short[] GetVersions() 
        {
            EAPSIMAttribute versionListAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_VERSION_LIST);
            if (versionListAttribute == null) return null;
            byte[] versionList = versionListAttribute.GetValue(true);
            short[] ret = new short[versionList.Length/2];
            byte[] version = new byte[2];
            for (int i = 0; i < versionList.Length; i = i + 2)
            {
                ret[i] = BitConverter.ToInt16(versionList, i);
            } 
            return ret;
        }
    }

    /// <summary>
    /// A class that represents the EAP-Response/SIM/Start message
    /// </summary>
    public class EAPResponseStart : EAPSIMMessage
    {
        public EAPResponseStart() : base(Code.Response, EAPType.SIM, EAPSIMMessage.Subtype.Start)
        {
        }
        
        /// <summary>
        /// Set the AT_IDENTITY attribute
        /// </summary>
        /// <param name="identity">The identity of the user</param>
        public void SetIdentity(byte[] identity)        
        {
            EAPSIMAttribute identityAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_IDENTITY, identity, true);
            AddAttribute(identityAttribute);
        }

        /// <summary>
        /// Get the AT_IDENTITY attribute
        /// </summary>
        /// <returns>The identity of the user</returns>
        public byte[] GetIdentity()                     
        {
            EAPSIMAttribute identityAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_IDENTITY);
            return identityAttribute.GetValue(true);
        }
        
        /// <summary>
        /// Set the AT_SELECTED_VERSION attribute
        /// </summary>
        /// <param name="version">A list of the selected versions</param>
        public void SetSelectedVersion(short version)   
        {
            byte[] value = BitConverter.GetBytes(version);
            EAPSIMAttribute versionAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_SELECTED_VERSION, value);
            AddAttribute(versionAttribute);
        }

        /// <summary>
        /// Get the AT_SELECTED_VERSION attribute
        /// </summary>
        /// <returns>The selected version</returns>
        public short GetSelectedVersion()               
        {
            EAPSIMAttribute versionAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_SELECTED_VERSION);
            byte[] value = versionAttribute.GetData();
            short selectedVersion = BitConverter.ToInt16(value, 0);
            return selectedVersion;
        }    
        
        /// <summary>
        /// Set the AT_NONCE_MT attribute
        /// </summary>
        /// <param name="nonce">The Nonce to set</param>
        public void SetNonce(byte[] nonce)              
        {
            if (nonce.Length != 16) throw new ArgumentException("proposal needs to be 16 bytes, was " + nonce.Length);
            EAPSIMAttribute nonceAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_NONCE_MT, nonce, false);
            AddAttribute(nonceAttribute);
        }
        
        /// <summary>
        /// Get the AT_NONCE_MT attribute
        /// </summary>
        /// <returns>The value of the Nonce</returns>
        public byte[] GetNonce()                        
        {
            EAPSIMAttribute nonceAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_NONCE_MT);
            return nonceAttribute.GetValue(false);
        }

        /// <summary>
        ///Generate and set a random AT_NONCE_MT attribute 
        /// </summary>
        /// <returns>The Nonce that was generated</returns>
        public byte[] GenerateNonce()                   
        {
            Random rand = new Random();
            byte[] nonce = new byte[16];
            rand.NextBytes(nonce);
            SetNonce(nonce);
            return nonce;
        }
    }

    /// <summary>
    /// A class for representing EAP-Request/SIM/Challenge messages
    /// </summary>
    public class EAPRequestChallenge : EAPSIMMessage    // Create an EAP-Request/SIM/Challenge message
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(EAPRequestChallenge));
        public EAPRequestChallenge() : base(Code.Request, EAPType.SIM, Subtype.Challenge) {}

        /// <summary>
        /// Set the AT_RAND attribute
        /// </summary>
        /// <param name="rands">The RAND(s) to set</param>
        public void SetRAND(byte[] rands)               
        {
            if (rands.Length % 16 != 0) throw new ArgumentException("Need n*16 bytes of rands, got " + rands.Length);
            EAPSIMAttribute randAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_RAND,rands,false );
            AddAttribute(randAttribute);
        }

        /// <summary>
        /// Return the AT_RAND attribute
        /// </summary>
        /// <returns>The RAND(s) that was sent</returns>
        public byte[] GetRand()                         
        {
            EAPSIMAttribute randAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_RAND);
            return randAttribute.GetValue(false);
        }

        /// <summary>
        /// Set the AT_NEXT_PSEUDONYM attribute
        /// </summary>
        /// <param name="np">The pseudonym to set</param>
        public void SetNextPseudonym(byte[] np)         
        {
            EAPSIMAttribute pseudonymAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_NEXT_PSEUDONYM, np, true);
            AddAttribute(pseudonymAttribute);
        }
        
        /// <summary>
        /// Get the AT_NEXT_PSEUDONYM attribute
        /// </summary>
        /// <returns>The next_pseudonym value</returns>
        public byte[] GetNextPseudonym()                
        {
            EAPSIMAttribute pseudonymAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_NEXT_PSEUDONYM);
            return pseudonymAttribute.GetValue(true);
        }
        
        /// <summary>
        /// Set the AT_NEXT_REAUTH_ID attribute
        /// </summary>
        /// <param name="id"></param>
        public void SetNextReauthID(byte[] id)          
        {
            EAPSIMAttribute reauthAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_NEXT_REAUTH_ID,id,true);
            AddAttribute(reauthAttribute);
        }
        
        /// <summary>
        /// Get the AT_NEXT_REAUTH_ID attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetNextReauthID()                 
        {
            EAPSIMAttribute reauthAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_NEXT_REAUTH_ID);
            return reauthAttribute.GetValue(true); 
        }
        
        /// <summary>
        /// Set the AT_IV attribute
        /// </summary>
        /// <param name="iv"></param>
        public void SetIV(byte[] iv)                    
        {
            EAPSIMAttribute ivAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_IV, iv, false);
            AddAttribute(ivAttribute);
        }
        
        /// <summary>
        /// Get the AT_IV attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetIV()                           
        {
            EAPSIMAttribute reauthAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_IV);
            return reauthAttribute.GetValue(false);
        }
        
        /// <summary>
        /// Set the AT_ENCR_DATA attribute
        /// </summary>
        /// <param name="ecd"></param>
        public void SetEncryptedData(byte[] ecd)        
        {
            EAPSIMAttribute ecdAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_ENCR_DATA, ecd, false);
            AddAttribute(ecdAttribute);
        }
        
        /// <summary>
        /// Get the AT_ENCR_DATA attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetEncryptedData()                
        {
            EAPSIMAttribute dataAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_ENCR_DATA);
            return dataAttribute.GetValue(true);

        }
        
        /// <summary>
        /// Set the AT_PADDING attribute
        /// </summary>
        /// <param name="size"></param>
        public void SetPadding(int size)                
        {
            EAPSIMAttribute paddingAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_PADDING, new byte[size]);
            AddAttribute(paddingAttribute);
        }
        
        /// <summary>
        /// Set the AT_RESULT_IND attribute
        /// </summary>
        public void SetResultInd()                      
        {
            EAPSIMAttribute resultIndAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_RESULT_IND, new byte[0], false);
            AddAttribute(resultIndAttribute);
        }
        
        /// <summary>
        /// Get the AT_RESULT_IND attribute
        /// </summary>
        /// <returns></returns>
        public bool GetResultInd()                      
        {
            if (GetAttribute(EAPSIMAttribute.AttributeType.AT_RESULT_IND) != null) return true;
            else return false;
        }
        
        /// <summary>
        /// Set the AT_MAC attribute
        /// </summary>
        /// <param name="mac"></param>
        public void SetMAC(byte[] mac)                  
        {
            EAPSIMAttribute macAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_MAC, mac, false);
            AddAttribute(macAttribute);
        }

        /// <summary>
        /// Generate the MAC of the message using HMAC-SHA1, setting itself to 0s before calculating. 
        /// And adding itself as an attribute afterwards. Use it after adding all the other attributes that should
        /// be included in the message.
        /// </summary>
        /// <param name="nonce">The nonce to include in the MAC</param>
        /// <param name="masterKey">The key for the HMACSHA1 function</param>
        public void GenerateMAC(byte[] nonce, byte[] masterKey)
        {
            byte[] mac = new byte[20];
            SetMAC(mac);
            //Find the bytes to calculate the MAC over
            byte[] eapBytes = GetBytes();
            byte[] bytes = new byte[eapBytes.Length+nonce.Length];
            Array.Copy(eapBytes,0,bytes,0,eapBytes.Length);
            Array.Copy(nonce, 0, bytes, eapBytes.Length, nonce.Length);
            HMACSHA1 sha = new HMACSHA1(masterKey);
            mac = sha.ComputeHash(bytes);
            UpdateMAC(mac); //Set the new calculated MAC value
        }

        /// <summary>
        /// Verify the MAC value of the message. Setting itself to 0s before calculating
        /// </summary>
        /// <param name="nonce">The nonce to include in the MAC</param>
        /// <param name="masterKey">The key for the HMACSHA1 function</param>
        /// <returns></returns>
        public bool VerifyMAC(byte[] nonce, byte[] masterKey)
        {
            byte[] mac = this.GetMAC();
            this.UpdateMAC(new byte[20]);

            byte[] eapBytes = GetBytes();
            byte[] bytes = new byte[eapBytes.Length + nonce.Length];
            Array.Copy(eapBytes, 0, bytes, 0, eapBytes.Length);
            Array.Copy(nonce, 0, bytes, eapBytes.Length, nonce.Length);
            HMACSHA1 sha = new HMACSHA1(masterKey);
            byte[] mac2 = sha.ComputeHash(bytes);

            if (mac.Length != mac2.Length) return false; //This shouldn't happen
            for (int i = 0; i < mac.Length; i++)
            {
                if (mac[i] != mac2[i]) return false;
            }
            return true;
        }

        /// <summary>
        /// Get the AT_MAC attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetMAC()                          
        {
            EAPSIMAttribute macAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_MAC);
            return macAttribute.GetValue(false);
        }

        /// <summary>
        /// Update the mac attribute of the message
        /// </summary>
        /// <param name="mac"></param>
        public void UpdateMAC(byte[] mac)
        {
            EAPSIMAttribute macAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_MAC, mac, false);
            UpdateAttribute(macAttribute);
        }

    }
    
    /// <summary>
    /// Represents the EAP-Response/SIM/Challenge message
    /// </summary>
    public class EAPResponseChallenge : EAPSIMMessage   
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(EAPResponseChallenge));
        public EAPResponseChallenge() : base(Code.Response, EAPType.SIM, Subtype.Challenge) {}

        /// <summary>
        /// Set the AT_RESULT_IND attribute
        /// </summary>
        public void SetResultInd()                      
        {
            EAPSIMAttribute resultIndAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_RESULT_IND, new byte[0], false);
            AddAttribute(resultIndAttribute);
        }
        
        /// <summary>
        /// Get the AT_RESULT_IND attribute
        /// </summary>
        /// <returns></returns>
        public bool GetResultInd()                      
        {
            if (GetAttribute(EAPSIMAttribute.AttributeType.AT_RESULT_IND) != null) return true;
            else return false;
        }
        
        /// <summary>
        /// Set the AT_MAC attribute
        /// </summary>
        /// <param name="mac"></param>
        public void SetMAC(byte[] mac)                  
        {
            EAPSIMAttribute macAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_MAC, mac, false);
            AddAttribute(macAttribute);
        }
        
        /// <summary>
        /// Get the AT_MAC attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetMAC()                          
        {
            EAPSIMAttribute macAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_MAC);
            return macAttribute.GetValue(false);
        }

        /// <summary>
        /// Verify the MAC value of the message. Setting itself to 0s before calculating
        /// </summary>
        /// <param name="nonce">The nonce to include in the MAC</param>
        /// <param name="masterKey">The key for the HMACSHA1 function</param>
        /// <returns></returns>
        public bool VerifyMAC(byte[] nonce, byte[] masterKey)
        {
            byte[] mac = this.GetMAC();
            this.UpdateMAC(new byte[20]);

            byte[] eapBytes = GetBytes();
            byte[] bytes = new byte[eapBytes.Length + nonce.Length];
            Array.Copy(eapBytes, 0, bytes, 0, eapBytes.Length);
            Array.Copy(nonce, 0, bytes, eapBytes.Length, nonce.Length);
            HMACSHA1 sha = new HMACSHA1(masterKey);
            byte[] mac2 = sha.ComputeHash(bytes);

            if (mac.Length != mac2.Length) return false; //This shouldn't happen
            for (int i = 0; i < mac.Length;i++)
            {
                if (mac[i] != mac2[i]) return false;
            }
            return true;                
        }

        /// <summary>
        /// Generate the MAC of the message using HMAC-SHA1, setting itself to 0s before calculating. 
        /// And adding itself as an attribute afterwards. Use it after adding all the other attributes that should
        /// be included in the message.
        /// </summary>
        /// <param name="nonce">The nonce to include in the MAC</param>
        /// <param name="masterKey">The key for the HMACSHA1 function</param>
        public void GenerateMAC(byte[] nonce, byte[] masterKey)
        {
            byte[] mac = new byte[20];
            SetMAC(mac);
            //Find the bytes to calculate the MAC over
            byte[] eapBytes = GetBytes();
            byte[] bytes = new byte[eapBytes.Length + nonce.Length];
            Array.Copy(eapBytes, 0, bytes, 0, eapBytes.Length);
            Array.Copy(nonce, 0, bytes, eapBytes.Length, nonce.Length);
            HMACSHA1 sha = new HMACSHA1(masterKey);
            mac = sha.ComputeHash(bytes);
  
            UpdateMAC(mac); //Set the new calculated MAC value
        }

        /// <summary>
        /// Update the MAC attribute
        /// </summary>
        /// <param name="mac"></param>
        public void UpdateMAC(byte[] mac)
        {
            EAPSIMAttribute macAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_MAC, mac, false);
            UpdateAttribute(macAttribute);
        }
    }
    
    /// <summary>
    /// Representation of an EAP-Request/SIM/Notification message
    /// </summary>
    public class EAPRequestNotification : EAPSIMMessage 
    {
        public EAPRequestNotification() : base(Code.Request, EAPType.SIM, Subtype.Notification) {}
        
        /// <summary>
        /// Set the AT_IV attribute
        /// </summary>
        /// <param name="iv"></param>
        public void SetIV(byte[] iv)                    
        {
            EAPSIMAttribute ivAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_IV, iv, false);
            AddAttribute(ivAttribute);
        }
        
        /// <summary>
        /// Get the AT_IV attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetIV()                           
        {
            EAPSIMAttribute reauthAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_IV);
            return reauthAttribute.GetValue(false);
        }
        
        /// <summary>
        /// Set the AT_ENCR_DATA attribute
        /// </summary>
        /// <param name="ecd"></param>
        public void SetEncryptedData(byte[] ecd)        
        {
            EAPSIMAttribute ecdAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_ENCR_DATA, ecd, false);
            AddAttribute(ecdAttribute);
        }
        
        /// <summary>
        /// Get the AT_ENCR_DATA attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetEncryptedData()                
        {
            EAPSIMAttribute dataAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_ENCR_DATA);
            return dataAttribute.GetValue(true);

        }
        
        /// <summary>
        /// Set the AT_PADDING attribute
        /// </summary>
        /// <param name="size"></param>
        public void SetPadding(int size)                
        {
            EAPSIMAttribute paddingAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_PADDING, new byte[size]);
            AddAttribute(paddingAttribute);
        }
        
        /// <summary>
        /// Set the AT_MAC attribute
        /// </summary>
        /// <param name="mac"></param>
        public void SetMAC(byte[] mac)                  
        {
            EAPSIMAttribute macAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_MAC, mac, false);
            AddAttribute(macAttribute);
        }
        
        /// <summary>
        /// Get the AT_MAC attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetMAC()                          
        {
            EAPSIMAttribute macAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_MAC);
            return macAttribute.GetValue(false);
        }
        
        /// <summary>
        /// Set the AT_COUNTER attribute
        /// </summary>
        /// <param name="counter"></param>
        public void SetCounter(short counter)           
        {
            byte[] cnt = BitConverter.GetBytes(counter);
            EAPSIMAttribute cntAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_COUNTER, cnt);
            AddAttribute(cntAttribute);
        }
        
        /// <summary>
        /// Get the AT_COUNTER attribute
        /// </summary>
        /// <returns></returns>
        public short GetCounter()                       
        {
            EAPSIMAttribute cntAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_COUNTER);
            short ret = BitConverter.ToInt16(cntAttribute.GetData(), 0);
            return ret;
        }
        
        /// <summary>
        /// Set the AT_NOTIFICATION attribute 
        /// </summary>
        /// <param name="nc"></param>
        public void SetNotification(byte[] nc)          
        {
            EAPSIMAttribute notificationAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_NOTIFICATION, nc);
            AddAttribute(notificationAttribute);
        }
        
        /// <summary>
        /// Get the AT_NOTIFICATION attribute 
        /// </summary>
        /// <returns></returns>
        public byte[] GetNotification()                 
        {
            EAPSIMAttribute notificationAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_NOTIFICATION);
            return notificationAttribute.GetData();
        }
    }
    
    /// <summary>
    /// Represents the EAP-Response/SIM/Notification message
    /// </summary>
    public class EAPResponseNotification : EAPSIMMessage 
    {
        public EAPResponseNotification() : base(Code.Response, EAPType.SIM, Subtype.Notification) { }
        
        /// <summary>
        /// Set the AT_CLIENT_ERROR_CODE attribute 
        /// </summary>
        /// <param name="ec"></param>
        public void SetErrorCode(byte[] ec)             
        {
            EAPSIMAttribute errorAttribute = new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_CLIENT_ERROR_CODE, ec);
            AddAttribute(errorAttribute);
        }
        
        /// <summary>
        /// Get the AT_CLIENT_ERROR_CODE attribute 
        /// </summary>
        /// <returns></returns>
        public byte[] GetErrorCode()                    
        {
            EAPSIMAttribute errorAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_CLIENT_ERROR_CODE);
            return errorAttribute.GetData();
        }
    } 
    
    /// <summary>
    /// Represents an EAP-Request/SIM/KeyExchange message
    /// </summary>
    public class EAPRequestKeyExchange : EAPSIMMessage  
    {
        public EAPRequestKeyExchange() : base(Code.Request, EAPType.SIM, Subtype.KeyExchange){}

        /// <summary>
        /// Set the key attributes for the key exchange message
        /// </summary>
        /// <param name="dhKe">The Diffie-Hellman KeyEchange-bytesequence </param>
        /// <param name="dhp">The DHParameters object</param>
        public void SetKeyattributes(byte[] dhKe, DHParameters dhp)
        {   
            AddAttribute(new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_KEYPARAMG, dhp.G, true));
            AddAttribute(new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_KEYPARAMP, dhp.P, true));
            AddAttribute(new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_PUBLICKEY, dhKe, true));
        }

        /// <summary>
        /// Get the Diffie-Hellman Key parameters
        /// </summary>
        /// <returns>A DHParameters object containing the parameters of the attribute</returns>
        public DHParameters GetKeyParameters()
        {
            DHParameters dhp = new DHParameters();
            EAPSIMAttribute keyAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_KEYPARAMG);
            dhp.G = keyAttribute.GetValue(true);
            keyAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_KEYPARAMP);
            dhp.P = keyAttribute.GetValue(true);
            return dhp;
        }

        /// <summary>
        /// Get the AT_PUBLICKEY attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetKey()
        {
            EAPSIMAttribute keyAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_PUBLICKEY);
            return keyAttribute.GetValue(true);
        }
    }

    /// <summary>
    /// Represent an EAP-Response/SIM/KeyExchange message
    /// </summary>
    public class EAPResponseKeyExchange : EAPSIMMessage  
    {
        public EAPResponseKeyExchange() : base(Code.Response, EAPType.SIM, Subtype.KeyExchange) { }

        /// <summary>
        /// Set the AT_PUBLICKEY attribute
        /// </summary>
        /// <param name="dhKe"></param>
        public void SetKey(byte[] dhKe)
        {
            AddAttribute(new EAPSIMAttribute(EAPSIMAttribute.AttributeType.AT_PUBLICKEY, dhKe, true));
        }

        /// <summary>
        /// Get the AT_PUBLICKEY attribute
        /// </summary>
        /// <returns></returns>
        public byte[] GetKey()
        {
            EAPSIMAttribute keyAttribute = GetAttribute(EAPSIMAttribute.AttributeType.AT_PUBLICKEY);
            return keyAttribute.GetValue(true);
        }
    }
}
