using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Text;
using log4net;
using Utility;

namespace gsm.datasource
{
    /// <summary>
    /// A class representing the triplet source in a complete system. In other cases this would be the HLR of the gsm network,
    /// or the AAA server. Here we use a file on disk to store a set of triplets for the IMSIs
    /// </summary>
    public class TripletSource
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(TripletSource));
        private DataSet ds;
        /// <summary>
        /// Constructor for the TripletSource. Reads the XML Schema from the file "database/Triplets.xsd", and loads the 
        /// file "database/Triplets.xml" which should contain a set of valid IMSIs and triplets. If this file does not exist,
        /// the authentication server will reject authentication attempts from all IMSIs. 
        /// </summary>
        public TripletSource()
        {
            ds= new DataSet();
            FileStream finschema = new FileStream("database/Triplets.xsd", FileMode.Open, FileAccess.Read, FileShare.Read);
            ds.ReadXmlSchema(finschema);
            finschema.Close();

            //Try to read Triplets from disk
            try
            {
                FileStream findata =
                    new FileStream("database/Triplets.xml", FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
                ds.ReadXml(findata);
                findata.Close();
            }catch(FileNotFoundException)
            {
                log.Error("Error reading file \"database/Triplets.xml\" from disk. There are at the moment no valid IMSIs.");
            }
        }
        /// <summary>
        /// Add a triplet to the current working set of Triplets. Use Write() to write the triplets to disk after adding.
        /// </summary>
        /// <param name="imsi">IMSI of the triplet to add</param>
        /// <param name="rand">RAND value of the triplet</param>
        /// <param name="sres">SRES value of the triplet</param>
        /// <param name="kc">Kc value of the triplet</param>
        public void AddTriplet(byte[] imsi, byte[] rand, byte[] sres, byte[] kc)
        {
            string imsiString = HexEncoding.IMSIConvert(imsi);
            DataTable table = ds.Tables[0];
            DataRow triplet = table.NewRow();
            triplet["imsi"] = imsiString;
            triplet["rand"] = HexEncoding.ToString(rand);
            triplet["sres"] = HexEncoding.ToString(sres);
            triplet["kc"] = HexEncoding.ToString(kc);
            table.Rows.Add(triplet);
        }
        /// <summary>
        /// Writes the current set of IMSIS and triplets to the file "database/Triplets.xml" on disk. 
        /// </summary>
        public void Write()
        {
            FileStream fout = new FileStream("database/Triplets.xml", FileMode.OpenOrCreate, FileAccess.Write, FileShare.ReadWrite);
            ds.WriteXml(fout);
            fout.Close();
        }
        /// <summary>
        /// Fetches one random triplet from the current set of triplets. Does not generate new triplets.
        /// </summary>
        /// <param name="imsi">The IMSI of the triplet to fetch</param>
        /// <param name="rand">Return parameter - RAND</param>
        /// <param name="sres">Return parameter - SRES</param>
        /// <param name="kc">Return parameter - Kc</param>
        /// <returns>Returns true if the IMSI had any triplets</returns>
        public bool ReadTriplet(byte[] imsi, out byte[] rand, out byte[] sres, out byte[] kc)
        {
            string imsiString = HexEncoding.IMSIConvert(imsi);
            DataTable table = ds.Tables[0];
            DataRow[] triplets = table.Select("imsi=" + imsiString);
            
            //Find random triplet 
            Random random = new Random();
            DataRow triplet = triplets[random.Next(0, triplets.Length)];

            //A place to put the discarded value, should be 0
            int discarded;
            rand = HexEncoding.GetBytes((string) triplet["rand"], out discarded);
            sres = HexEncoding.GetBytes((string)triplet["sres"], out discarded);
            kc = HexEncoding.GetBytes((string)triplet["kc"], out discarded);
            return true;
        }

        /// <summary>
        /// A function used to fake the SRES calculation function, for use by dummy SIM card. Instead of 
        /// calculating the SRES with the GSM algorithm of the SIM, instead read from the triplet file.
        /// The triplet file here should (must) contain the same triplets as in the authenticator triplet file.
        /// </summary>
        /// <param name="imsi">IMSI of the triplet to look up</param>
        /// <param name="rand">RAND of the triplet to look up</param>
        /// <param name="sres">Return SRES value "calculated" by the GSM algorithm </param>
        /// <param name="kc">Return Kc value as "calculated" by the GSM algorithm</param>
        /// <returns>Returns whether the imsi+rand triplet was found.</returns>
        public bool GetSRES(byte[] imsi, byte[] rand, out byte[] sres, out byte[] kc)
        {
            sres = new byte[] { };
            kc = new byte[] { };
            string imsiString = HexEncoding.IMSIConvert(imsi);
            DataTable table = ds.Tables[0];
            DataRow[] triplets = table.Select("imsi='" + imsiString+"' and rand='"+HexEncoding.ToString(rand)+"'");
            if (triplets.Length == 0) return false;
            
            //Return the sres and kc values
            int discarded;
            sres = HexEncoding.GetBytes((string)triplets[0]["sres"], out discarded);
            kc = HexEncoding.GetBytes((string)triplets[0]["kc"], out discarded);
            return true;
        }
        
        /// <summary>
        /// A function for validating an IMSI, based on the set of existing triplets. If the IMSI does not 
        /// have any triplets, it is considered invalid/unknown.
        /// </summary>
        /// <param name="imsi">The IMSI to validate</param>
        /// <returns>True if the IMSI has any associated triplets</returns>
        public bool ValidateIMSI(byte[] imsi)
        {
            char[] cImsi = new char[imsi.Length];
            for (int i = 0; i < imsi.Length; i++)
            {
                cImsi[i] = imsi[i].ToString()[0];
            }
            string imsiString = new string(cImsi);
            DataTable table = ds.Tables[0];
            DataRow[] triplets = table.Select("imsi=" + imsiString);
            if (triplets.Length > 0) return true;
            else return false;
        }
    }
}
