using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using eap;
using log4net;
using vpnsim.supplicant.applicationSupplicant.eapCommunicator;

namespace vpnsim.supplicant.applicationSupplicant.core
{
    /// <summary>
    /// Class for communicating with the authenticator
    /// </summary>
    public class Communicator
    {
        public Thread communicatorThread;
        private SimClient simSupplicant;
        private Supplicant supplicant;
        private static readonly ILog log = LogManager.GetLogger(typeof(Communicator));
        private IPEndPoint authenticatorAddress;

        /// <summary>
        /// Constructor for the Communicator class
        /// </summary>
        /// <param name="supplicant"></param>
        public Communicator(Supplicant supplicant)
        {
            this.supplicant = supplicant;
        }

        /// <summary>
        /// Constructor overload for the communicator class, allows specifying the authenticator
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <param name="port"></param>
        /// <param name="supplicant"></param>
        public Communicator(string ipAddress, int port, Supplicant supplicant) : this(supplicant)
        {
            authenticatorAddress = new IPEndPoint(IPAddress.Parse(ipAddress),port);
        }

        /// <summary>
        /// Get the IP address of the authentication supplicant
        /// </summary>
        /// <returns></returns>
        public string GetAuthSupplicantAddress()
        {
            return authenticatorAddress.Address.ToString();
        }

        /// <summary>
        /// Get the authentication supplicant port number
        /// </summary>
        /// <returns></returns>
        public int GetAuthSupplicantPort()
        {
            return authenticatorAddress.Port;
        }

        /// <summary>
        /// Sets a new authentication supplicant
        /// </summary>
        /// <param name="address"></param>
        /// <param name="port"></param>
        /// <returns></returns>
        public bool ChangeAuthSupplicant(string address, string port)
        {
            
            IPAddress newIP;
            int newPort;
            bool ipResult = IPAddress.TryParse(address,out newIP);
            bool portResult = Int32.TryParse(port, out newPort);
            if (ipResult && portResult)
            {
                IPEndPoint newAddress = new IPEndPoint(newIP, newPort);
                if(newAddress.Equals(authenticatorAddress)) return true;
                authenticatorAddress = newAddress;
                log.Debug("Authenticator address changed to "+address+":"+port);
                return true;
            }
            else return false;
        }

        /// <summary>
        /// The main authentication method. Triggers the thread running communication with the authenticator
        /// </summary>
        private void RunAuthentication()
        {
            bool ongoing = true;

            EAPHandler eapHandler = new EAPHandler(simSupplicant);
            EapClient eapClient = new EapClient(authenticatorAddress);
            EAPMessage response = new EAPSuccessFailureMessage(Code.Failure);
            EAPResponseStart startMessage = new EAPResponseStart();
            byte[] imsi;
            simSupplicant.GetIMSI(out imsi);
            startMessage.SetIdentity(imsi);

            startMessage.SetSelectedVersion(1);

            byte[] nonce = startMessage.GenerateNonce();

            eapHandler.imsi = imsi; //Store the imsi for later use
            eapHandler.selectedVersion = 1;
            eapHandler.versionList = new short[]{1};
            eapHandler.nonce = nonce;
            
            EAPMessage message = (EAPMessage) startMessage;

            while(ongoing) //While the communication with the authenticator is ongoing
            {
                log.Info("Sending " + message.ToString());
                response = eapClient.Send(message);
                if(response==null)
                {
                    
                    supplicant.AuthenticationComplete(false);
                    return;
                }
                log.Info("Received " + response.ToString());
                //Abort if we have a failure (or success)

                message = eapHandler.HandleMessage(response);
                if (response.code.Equals(Code.Failure) || response.code.Equals(Code.Success)) ongoing = false;
            }
            supplicant.AuthenticationComplete(response.code.Equals(Code.Success));
            
        }

        /// <summary>
        /// Stop the authenticator thread
        /// </summary>
        public void Stop()
        {
            if(communicatorThread!=null)communicatorThread.Abort();
        }

        /// <summary>
        /// Starts running the authentication mechanism
        /// </summary>
        /// <param name="simSupplicant">SimSupplicant object to use for communication with the SIM</param>
        /// <returns></returns>
        public bool Authenticate(SimClient simSupplicant)
        {
            this.simSupplicant = simSupplicant;
            try
            {
                communicatorThread = new Thread(new ThreadStart(RunAuthentication));
                communicatorThread.Start();
                return true;
            }
            catch (Exception)
            {
                communicatorThread.Abort();
                return false;
            }
        }
    }
}
