using System;
using System.Security.Cryptography;
using System.Text;
using System.Windows.Forms;
using eap;
using gsm.datasource;
using log4net;
using Utility;
using vpnsim.supplicant;
using vpnsim.supplicant.applicationSupplicant.eapCommunicator;
using vpnsim.supplicant.simSupplicant;
using vpnsim.supplicant.simSupplicant.bluetooth;
using vpnsim.supplicant.simSupplicant.smartCard;

namespace vpnsim.supplicant.applicationSupplicant.core
{

    /// <summary>
    /// Class for setting parameters for SIM communication
    /// </summary>
    public class SimClient : ISimFunctions
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(SimClient));
        private char[] pinCode;
        public bool connected, chvVerified = false;
        private SimInterface selectedInterface;
        public byte currentInterfaceNumber = 9;
        public string scInterfaceName;
        public string btInterfaceAddress;
        private Supplicant mySupplicant;

        /// <summary>
        /// Creates a new instance of the SimClient class
        /// </summary>
        /// <param name="mySupplicant">The parent Supplicant object</param>
        public SimClient(Supplicant mySupplicant)
        {
            this.mySupplicant = mySupplicant;
            pinCode = new char[4];
        }

        /// <summary>
        /// Overload constructor for the SimClient class.
        /// </summary>
        /// <param name="sInterface">The SimInterface to use</param>
        /// <param name="mySupplicant">The parent Supplicant object</param>
        public SimClient(SimInterface sInterface, Supplicant mySupplicant) : this(mySupplicant)
        {
            
            this.selectedInterface = sInterface;
            
        }
        /// <summary>
        /// Register a key in the authentication server database
        /// </summary>
        /// <param name="imsi">The username to register</param>
        /// <param name="key">The key to register</param>
        public void RegisterKey(byte[] imsi, byte[] key)
        {
            string imsiString = HexEncoding.IMSIConvert(imsi);
            mySupplicant.AuthenticateUser(imsiString, key);
        }


        /// <summary>
        /// Is a valid interface set?
        /// </summary>
        /// <returns></returns>
        public bool IsValidInterface()
        {
            if (selectedInterface == null || !selectedInterface.IsReady()) return false;
            else return true;
        }
        
        /// <summary>
        /// Set a new interface
        /// </summary>
        /// <param name="interfaceNumber">The interface number to use</param>
        public void ChangeInterface(byte interfaceNumber)
        {
            //No switch, do nothing.
            if (interfaceNumber.Equals(currentInterfaceNumber)) return;
            //Disconnect the existing interface
            Disconnect();
            switch (interfaceNumber)
            {
                case 0x00: //Smart card reader
                    selectedInterface = new SmartCardInterface(scInterfaceName);
                    log.Debug("SIM Interface changed to Smart Card Reader - " + scInterfaceName);
                    break;
                case 0x01: //Bluetooth interface
                    selectedInterface = new BluetoothInterface(btInterfaceAddress);
                    log.Debug("SIM Interface changed to Bluetooth - " + btInterfaceAddress);
                    break;
                case 0x02: //Dummy Interface
                    selectedInterface = new DummyInterface();
                    log.Debug("SIM Interface changed to DummyInterface");
                    break;
            }
            connected = chvVerified = false;
            currentInterfaceNumber = interfaceNumber;
            
        }

        /// <summary>
        /// Set the PIN code of the SIM
        /// </summary>
        /// <param name="pin">New pin code</param>
        /// <returns>Whether the PIN code format was ok</returns>
        public bool ChangePinCode(string pin)
        {
            if(pin.Length>8) throw new ArgumentException("Pincode too long, expected <8 bytes, got " + pin.Length);
            //Test if the pin code was a numeric string
            int intTest;
            bool isNumeric = Int32.TryParse(pin, out intTest);
            if(!isNumeric) return false;
            pinCode = pin.ToCharArray();
            return true;
        }

        /// <summary>
        /// Get the currently set PIN code
        /// </summary>
        /// <returns></returns>
        public string GetPinCode()
        {
            char[] cPin = new char[4];
            for (int i = 0; i < pinCode.Length; i++)
            {
                cPin[i] = (char) pinCode[i];
            }
            return new string(cPin);
        }

        /// <summary>
        /// Connect to the current interface
        /// </summary>
        /// <returns>If the connection attempt was successful</returns>
        public bool Connect()
        {
            if (connected) return true;
            connected = selectedInterface.Connect();
            return connected;
        }

        /// <summary>
        /// Disconnect from the current interface
        /// </summary>
        public void Disconnect()
        {
            if (!connected) return;
            selectedInterface.Disconnect();
            connected = chvVerified = false;
        }

        /// <summary>
        /// Verify the PIN code against the current interface, using the currently set pin code
        /// </summary>
        /// <returns></returns>
        public bool VerifyCHV()
        {
            return VerifyCHV(pinCode);
        }
        
        /// <summary>
        /// Overload for VerifyCHV(). Verify the given PIN against the current interface
        /// </summary>
        /// <param name="chv">The PIN code to use</param>
        /// <returns></returns>
        public bool VerifyCHV(char[] chv)
        {
            if(!connected) Connect();
            if (!connected) return false;
            if (chvVerified) return true;
            chvVerified = selectedInterface.VerifyCHV(chv);
            log.Debug("Verify CHV - " + selectedInterface.GetResponse().ToString());
            return !selectedInterface.GetResponse().error;
        }

        /// <summary>
        /// Get the IMSI of the SIM inserted into the selected interface
        /// </summary>
        /// <param name="imsi">The IMSI value read</param>
        /// <returns>If the IMSI was read</returns>
        public bool GetIMSI(out byte[] imsi)
        {
            imsi = new byte[15];
           // if (chvVerified)
            {
                selectedInterface.SelectFile(SIM.DedicatedFile.GSM);
                selectedInterface.SelectFile(SIM.ElementaryFile.IMSI);

                bool result = selectedInterface.ReadBinary((byte)9); //IMSI length 9, for some reason bluetooth doesn't give
                //byte[] imsires = selectedInterface.GetResponse().responseData;
                //byte imsiLength = imsires[3];
                //bool result = selectedInterface.ReadBinary((byte)imsiLength); //IMSI length 9
                if (result)
                {
                    log.Debug("Read IMSI - " + selectedInterface.GetResponse().ToString());
                    byte[] response = selectedInterface.GetResponse().DATA;
                    byte length = (byte)(response[0]+1);
                    byte parity = (byte) (response[1] & 0x08);

                    imsi[0] = (byte) ((response[1] & 0xF0) >> 4);
                    int index = 1;
                    for (int i = 2; i < length; i++)
                    {
                        imsi[index++] = (byte) (response[i] & 0x0F);
                        imsi[index++] = (byte) ((response[i] & 0xF0) >> 4);
                    }
                   
                    //log.Debug("IMSI: " + HexEncoding.IMSIConvert(imsi));
                    return true;
                }
                log.Error("Unable to read imsi, cannot read EF " + selectedInterface.GetResponse().ToString());
                return false;
            }
        }

        /// <summary>
        /// Not implemented/used here
        /// </summary>
        /// <param name="imsi"></param>
        /// <returns></returns>
        public bool ValidateIMSI(byte[] imsi)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Not implemented/used here
        /// </summary>
        /// <param name="imsi"></param>
        /// <param name="i"></param>
        /// <param name="sres"></param>
        /// <param name="Kc"></param>
        /// <returns></returns>
        public byte[] GetRANDs(byte[] imsi, int i, out byte[] sres, out byte[] Kc)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Calculates the SRES
        /// </summary>
        /// <param name="randbytes">Array of RAND bytes to calculate SRES for</param>
        /// <param name="sres">The returned SRES value</param>
        /// <param name="Kc">The returned Kc value</param>
        /// <returns>If the calculation was successful</returns>
        public bool CalculateSRES(byte[] randbytes, out byte[] sres, out byte[] Kc)
        {
            if (!chvVerified)
            {
                Kc = sres = new byte[]{};
                return false;
            }

            selectedInterface.SelectFile(SIM.DedicatedFile.GSM);
            int rands = randbytes.Length/16;

            sres = new byte[4*rands];
            Kc = new byte[8*rands];
            byte[] rand = new byte[16];
            for (int i = 0; i < rands; i++)
            {
                Array.Copy(randbytes, i*16, rand, 0, 16);
                bool result = selectedInterface.RunGSMAlgorithm(rand);
                if (selectedInterface.GetResponse().error) return false;
                
                Array.Copy(selectedInterface.GetResponse().DATA, 0, sres, 4*i, 4);
                Array.Copy(selectedInterface.GetResponse().DATA, 4, Kc, 8*i, 8);
                
            }
            return true;
        }

        /// <summary>
        /// Calculate the master key for this connection
        /// </summary>
        /// <param name="imsi"></param>
        /// <param name="kc"></param>
        /// <param name="nonce"></param>
        /// <param name="verList"></param>
        /// <param name="selectedVersion"></param>
        /// <returns>The master key</returns>
        public byte[] GetMasterKey(byte[] imsi, byte[] kc, byte[] nonce, short[] verList, short selectedVersion)
        {
            byte[] data = new byte[imsi.Length + kc.Length + nonce.Length + verList.Length*2 + 2];
            Array.Copy(imsi, 0, data, 0, imsi.Length);
            Array.Copy(kc, 0, data, imsi.Length, kc.Length);
            Array.Copy(nonce, 0, data, imsi.Length+kc.Length, nonce.Length);
            byte[] verListBytes = new byte[verList.Length*2];
            for(int i = 0;i<verList.Length;i++)
            {
                verListBytes[2*i] = (byte)((verList[i]&0xFF00) >> 8);
                verListBytes[2*i+1] = (byte)((verList[i]&0x00FF));
            }
            byte[] selVerBytes = new byte[2];
            selVerBytes[0] = (byte) ((selectedVersion & 0xFF00) >> 8);
            selVerBytes[1] = (byte) ((selectedVersion & 0x00FF));
            Array.Copy(verListBytes, 0, data, imsi.Length+kc.Length+nonce.Length, verListBytes.Length);
            Array.Copy(selVerBytes, 0, data, imsi.Length+kc.Length+nonce.Length+verListBytes.Length, 2);
            SHA1 sha = new SHA1CryptoServiceProvider();
            byte[] result = sha.ComputeHash(data);
            return result;
        }

        /// <summary>
        /// Debug function for generating a random set of triplets and write them to file
        /// </summary>
        /// <param name="number">The number of triplets to generate</param>
        /// <returns>Whether the triplets could be generated and written</returns>
        public bool GenerateTriplet(int number)
        {
            if (!chvVerified) return false;
            byte[] imsi;
            bool gotImsi = GetIMSI(out imsi);
            if(!gotImsi)
            {
                log.Debug("Problem retrieving imsi, error");
                return false;
            }
            Random random = new Random();
            byte[] rand = new byte[16];
            TripletSource ts = new TripletSource();
            for (int i = 0; i < number; i++)
            {

                random.NextBytes(rand);
                log.Debug("Generating random triplet for IMSI " + HexEncoding.IMSIConvert(imsi));
                byte[] sres, kc;
                bool res = CalculateSRES(rand, out sres, out kc);
                if (!res)
                {
                    log.Debug("Problem calculating sres, error");
                    return false ;
                }
                
                ts.AddTriplet(imsi, rand, sres, kc);
            }
            ts.Write();
            MessageBox.Show("Triplets written to file:\r\n"+Application.StartupPath+"\\database\\Triplets.xml", "Triplets Generated");
            return true;
        }
    }
}