using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Cryptography;
using System.Windows.Forms;
using log4net;
using vpnsim.supplicant;
using vpnsim.supplicant.applicationSupplicant.userInterface;
using vpnsim.supplicant.simSupplicant.bluetooth;

namespace vpnsim.supplicant.applicationSupplicant.core
{
    /// <summary>
    /// The main class for this application
    /// </summary>
    public class Supplicant
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(Supplicant));

        private static DebugLog logWindow;

        [STAThread] 
        static void Main()
        {
                
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            logWindow = new DebugLog();

            SupplicantForm form = new SupplicantForm(logWindow);
            Supplicant supplicant = new Supplicant(form);
            form.SetSupplicant(supplicant);
            Application.Run();
            
            
        }

        private Communicator communicator;
        private SimClient simSupplicant;
        private SupplicantForm window;
        private VPNManagementClient vpnMgmtClient;
        public User authenticatedUser;
        public bool authenticating;
        public bool running;

        /// <summary>
        /// Constructor for the SupplicantClass
        /// </summary>
        /// <param name="window">The main settings window</param>
        public Supplicant(SupplicantForm window)
        {
            this.window = window;

            simSupplicant = new SimClient(this);
            simSupplicant.btInterfaceAddress = Properties.Settings.Default.btInterfaceAddress;
            simSupplicant.scInterfaceName = Properties.Settings.Default.scInterfaceName;
            
            string asAddress = Properties.Settings.Default.authSupplicantAddress;
            int asPort = Properties.Settings.Default.authSupplicantPort;
            communicator = new Communicator(this);

            if (Properties.Settings.Default.settingsSet)
            {            
                //Start with old settings
                window.DisplayWindow(false);
                ChangeAuthSupplicant(asAddress, asPort.ToString());
                ChangePinCode(Properties.Settings.Default.pinCode);
                bool interfaceOk = ChangeInterface(Properties.Settings.Default.simCommunication);
                if(interfaceOk) Ready();
                else
                {
                    Properties.Settings.Default.settingsSet = false;
                    window.DisplayWindow(true);
                }
            }else
            {
                window.DisplayWindow(true);
            }
        }

        /// <summary>
        /// Get the SimSupplicant object of the supplicant
        /// </summary>
        /// <returns></returns>
        public SimClient GetSimSupplicant()
        {
            return simSupplicant;
        }

        /// <summary>
        /// Get the communicator object of the supplicant
        /// </summary>
        /// <returns></returns>
        public Communicator GetCommunicator()
        {
            return communicator;
        }

        /// <summary>
        /// Start the authentication mechanism
        /// </summary>
        public void Authenticate()
        {
            authenticating = true;
            if (!simSupplicant.chvVerified) return;
            communicator.Authenticate(simSupplicant);
        }

        /// <summary>
        /// Change the authentication supplicant address
        /// </summary>
        /// <param name="address"></param>
        /// <param name="port"></param>
        /// <returns></returns>
        public bool ChangeAuthSupplicant(string address, string port)
        {
            return communicator.ChangeAuthSupplicant(address, port);
        }

        /// <summary>
        /// Change the set PIN code
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        public bool ChangePinCode(string text)
        {
            return simSupplicant.ChangePinCode(text);
        }

        /// <summary>
        /// Change the current interface
        /// </summary>
        /// <param name="newInterface"></param>
        public bool ChangeInterface(byte newInterface)
        {
            simSupplicant.ChangeInterface(newInterface);
            if (!simSupplicant.IsValidInterface()) return false;
            if (!simSupplicant.Connect())
            {
                window.Error("There was an error connecting to the SIM Reader.");
                return false;
            }
            if(!simSupplicant.VerifyCHV())
            {
                window.Error("There was an error validating the PIN-code, please verify.");
                return false;
            }
            return true;
        }

        /// <summary>
        /// Write the current settings to the application settings
        /// </summary>
        public void SaveSettings()
        {
            //Save to application settings
            Properties.Settings.Default.pinCode = simSupplicant.GetPinCode();
            Properties.Settings.Default.authSupplicantAddress = communicator.GetAuthSupplicantAddress();
            Properties.Settings.Default.authSupplicantPort = communicator.GetAuthSupplicantPort();
            Properties.Settings.Default.btInterfaceAddress = simSupplicant.btInterfaceAddress;
            Properties.Settings.Default.scInterfaceName = simSupplicant.scInterfaceName;
            Properties.Settings.Default.simCommunication = simSupplicant.currentInterfaceNumber;
            Properties.Settings.Default.settingsSet = true;
        }

        /// <summary>
        /// A function that is called when the authentication mechanism is complete
        /// </summary>
        /// <param name="success">Specifies if the authentication was successful</param>
        public void AuthenticationComplete(bool success)
        {
            authenticating = false;
            string result = "Failed";
            if (success) result = "Successful";
            log.Info("Authentication Complete - "+result);
        }

        /// <summary>
        /// Set the status of the VPN Connection (for changing the tray icon)
        /// </summary>
        /// <param name="connected"></param>
        public void VPNConnected(bool connected)
        {
            window.setConnected(connected);
        }

        /// <summary>
        /// Generate random authentication triplets
        /// </summary>
        public void GenerateTriplet()
        {
            simSupplicant.GenerateTriplet(3);
        }

        /// <summary>
        /// Set the user as authenticated. 
        /// </summary>
        /// <param name="username">Username of the user</param>
        /// <param name="key">Key of the user</param>
        public void AuthenticateUser(string username, byte[] key)
        {
            MD5 md5 = new MD5CryptoServiceProvider();
            byte[] hash = md5.ComputeHash(key);

            string hexString = "";
            for (int i = 0; i < hash.Length; i++)
            {
                hexString += hash[i].ToString("X2");
            }

            window.setAuthenticated(true);
            this.authenticatedUser = new User(username, hexString.ToLower());
        }

        /// <summary>
        /// Set if the application supplicant is ready to start. When all the initialization is complete
        /// </summary>
        public void Ready()
        {
            if (!running)
            {
                vpnMgmtClient = new VPNManagementClient(this);
                vpnMgmtClient.Connect();
                running = true;
            }
        }

        /// <summary>
        /// Stop the application and exit
        /// </summary>
        public void Exit()
        {
            Properties.Settings.Default.Save();
            if(vpnMgmtClient!=null)vpnMgmtClient.Stop();
            communicator.Stop();
            Application.Exit();
        }
    }


    /// <summary>
    /// A class representing a user
    /// </summary>
    public class User
    {
        public string username;
        public string key;
        public User(string s, string key)
        {
            this.username = s;
            this.key = key;
        }
    }
}