using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using log4net;

namespace vpnsim.supplicant.applicationSupplicant.core
{
    using System;
    using System.Net;
    using System.Net.Sockets;
    using System.Text;

    /// <summary>
    /// A class for communicating with the VPN Management Interfaec
    /// </summary>
    public class VPNManagementClient
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(VPNManagementClient));
        private Thread vpnThread;
        private Socket server;
        private IPEndPoint endPoint;
        private bool connected = false;
        private bool tryConnect = false;
        private Supplicant supplicant;

        /// <summary>
        /// Constructor for the VPN Management Client
        /// </summary>
        /// <param name="supplicant">The parent supplicant</param>
        public VPNManagementClient(Supplicant supplicant)
        {
            this.supplicant = supplicant;
            endPoint = new IPEndPoint(IPAddress.Parse("127.0.0.1"), 4593);
        }

        /// <summary>
        /// Stop the thread
        /// </summary>
        public void Stop()
        {
            vpnThread.Abort();
        }

        /// <summary>
        /// Try connecting to the Management Interface
        /// </summary>
        /// <returns></returns>
        public bool Connect()
        {
            try
            {
                tryConnect = true;
                vpnThread = new Thread(new ThreadStart(TryConnect));
                vpnThread.Start();
                return true;
            }
            catch (Exception)
            {
                vpnThread.Abort();
                return false;
            }
        }

        private void TryConnect()
        {
            byte[] data = new byte[1024];
            while(tryConnect)
            {
                // Wait for VPN Client to start
                while (!connected)
                {
                    try
                    {
                        server = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                        server.Connect(endPoint);
                        connected = true;
                    }
                    catch (SocketException)
                    {
                        Thread.Sleep(100);
                    }
                }
                while(connected)
                {
                    bool poll = server.Poll(1000, SelectMode.SelectRead);;
                    if (poll) //Either the VPN Client has disconnected, or we have no more data to read
                    {
                        int recv = server.Receive(data);
                        if (recv == 0)
                        {
                            //Lost connection with VPN Client
                            log.Debug("Disconnected from VPN Client");
                            Disconnect();
                            continue;
                        }
                        else
                        {
                            //Yep, the VPN Client said something
                            string request = Encoding.ASCII.GetString(data, 0, recv);
                            string[] response;
                            bool answer = ParseRequest(request, out response);
                            if (answer)
                            {
                                log.Debug("Sending username/password to VPN Client");
                                foreach (string res in response) server.Send(Encoding.ASCII.GetBytes(res));
                            }
                        }
                    }else
                    {
                        //Nothing from the server
                        Thread.Sleep(1000);
                    }
                    if (!server.Connected) //The VPN Client has actively disconnected
                    {
                        log.Debug("Disconnected from VPN Client");
                        Disconnect();
                        continue;
                    }

                }
            }
        }
        private bool ParseRequest(string request, out string[] response)
        {
            Match m = Regex.Match(request, ">[A-Z]+");
            response = null;

            while(m.Success){

                switch (m.ToString())
                {
                    case ">PASSWORD":
                        log.Debug("Received username/password request from VPN Client");
                        return getResponse(out response);
                }
                m = m.NextMatch();
            }
            
            
            return false;
        }

        private bool getResponse(out string[] response)
        {
            if (supplicant.authenticatedUser == null)
            {
                supplicant.Authenticate();
            }
            while (supplicant.authenticatedUser == null)
            {
                //If this one is suddenly false, authentication failed, so send a sighup to the client
                if (!supplicant.authenticating)
                {
                    response = new string[]{"signal SIGHUP\r\n"};
                    return true;
                }
                Thread.Sleep(1000);
            }
            response =
                new string[]
                    {
                        "username \"Auth\" " + supplicant.authenticatedUser.username+"\r\n",
                        "password \"Auth\" " + supplicant.authenticatedUser.key+"\r\n"
                    };
            return true;
        }
        private void Disconnect()
        {
            connected = false;
            server.Close();
        }
    }
}
