using System;
using System.Drawing;
using System.ComponentModel;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using log4net;
using log4net.Config;
using log4net.Core;
//using log4net.spi;

namespace vpnsim.supplicant.applicationSupplicant.userInterface
{
    /// <summary>
    /// Debug Window for tailing log4net messages. Starts logging to interal buffer on create, not show.
    /// Closing window only hides it. To truly close it and stop the internal worker thread use Shutdown()
    /// </summary>
    public partial class DebugLog : Form
    {
        #region Static Members

        private static readonly ILog log = LogManager.GetLogger(typeof (DebugLog));
        #endregion

        #region Member Variables
        private bool shutdown = false;
        private bool logWatching = true;
        private log4net.Appender.MemoryAppender logger;
        private Thread logWatcher;
        private StringBuilder debugLog = new StringBuilder(32767);
        readonly object padlock = new object();

        private System.Windows.Forms.TextBox textBoxDebugLog;
        /// <summary>
        /// Required designer variable.
        /// </summary>
        private System.ComponentModel.Container components = null;

        #endregion

        public DebugLog()
        {
            //
            // Required for Windows Form Designer support
            //
            InitializeComponent();

            logger = new log4net.Appender.MemoryAppender();

            // Could use a fancier Configurator if you don't want to catch every message
            log4net.Config.BasicConfigurator.Configure(logger);

            // Since there are no events to catch logging, we dedicate
            // a thread to watching for logging "events"
            logWatcher = new Thread(new ThreadStart(LogWatcher));
            logWatcher.Start();

            this.Visible = false;

            this.Closing += new CancelEventHandler(DebugLog_Closing);
        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (components != null)
                {
                    components.Dispose();
                }
            }
            base.Dispose(disposing);
        }

        #region Windows Form Designer generated code
        /// <summary>
        /// Required method for Designer support - do not modify
        /// the contents of this method with the code editor.
        /// </summary>
        private void InitializeComponent()
        {
            this.textBoxDebugLog = new System.Windows.Forms.TextBox();
            this.SuspendLayout();
            // 
            // textBoxDebugLog
            // 
            this.textBoxDebugLog.Dock = System.Windows.Forms.DockStyle.Fill;
            this.textBoxDebugLog.Location = new System.Drawing.Point(0, 0);
            this.textBoxDebugLog.Multiline = true;
            this.textBoxDebugLog.Name = "textBoxDebugLog";
            this.textBoxDebugLog.ReadOnly = true;
            this.textBoxDebugLog.ScrollBars = System.Windows.Forms.ScrollBars.Vertical;
            this.textBoxDebugLog.Size = new System.Drawing.Size(745, 316);
            this.textBoxDebugLog.TabIndex = 15;
            // 
            // DebugLog
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(745, 316);
            this.Controls.Add(this.textBoxDebugLog);
            this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.SizableToolWindow;
            this.Name = "DebugLog";
            this.Text = "DebugLog";
            this.Load += new System.EventHandler(this.DebugLog_Load);
            this.ResumeLayout(false);
            this.PerformLayout();

        }
        #endregion

        private void DebugLog_Closing(object sender, CancelEventArgs e)
        {
            if (!shutdown)
            {
                // we only hide on normal close
                this.Hide();
                e.Cancel = true;
            }
        }

        /// <summary>
        /// Shutdown threads and call close so that it doesn't just hide the window
        /// </summary>
        public void Shutdown()
        {
            // set our shutdown flag
            shutdown = true;
            this.Close();
            // Gotta stop our logging thread
            logWatching = false;
            logWatcher.Join();
        }

        /// <summary>
        /// The logging worker which runs on it's own thread. Cannot write directly to form!
        /// </summary>
        private void LogWatcher()
        {
            // we loop until the Form is shut down
            while (logWatching)
            {
                // get all currently queued events
                LoggingEvent[] events = logger.GetEvents();

                if (events != null && events.Length > 0)
                {
                    // if there are events, we clear them from the logger,
                    // since we're done with them
                    logger.Clear();

                    // set up our sub-buffer for the current list of events
                    StringBuilder builder = new StringBuilder();
                    foreach (LoggingEvent ev in events)
                    {
                        // the line we want to log
                        builder.Append(ev.LoggerName + ": " + ev.RenderedMessage + "\r\n");
                    }
                    // Append to our internal buffer
                    debugLog.Append(builder);

                    if (debugLog.Length > 30000)
                    {
                        // buffer is getting large, let's trim it down some
                        debugLog.Remove(0, debugLog.ToString().IndexOf('\r', 5000) + 2);

                        if (this.IsHandleCreated)
                        {
                            // we got a window handle, so we can invoke the call to write out log to the window
                            this.BeginInvoke(new UpdateWholeLogDelegate(UpdateWholeLog), null);
                        }
                    }
                    else
                    {
                        // normal size log, we just append our new lines
                        if (this.IsHandleCreated)
                        {
                            // we got a window handle, so we can invoke the call to write out log to the window
                            this.BeginInvoke(new UpdateLogDelegate(UpdateLog), new object[] { builder.ToString() });
                        }
                    }
                }
                // nap for a while, don't need the events on the millisecond. 
                Thread.Sleep(500);
            }
        }

        private delegate void UpdateLogDelegate(string block);
        /// <summary>
        /// Update the log textbox in the form with a set of lines. This has to runson the form's thread
        /// so it should be called only via Invoke
        /// </summary>
        /// <param name="block"></param>
        private void UpdateLog(string block)
        {
            // using AppendText since that makes sure the TextBox stays
            // scrolled at the bottom
            textBoxDebugLog.AppendText(block);
        }

        private delegate void UpdateWholeLogDelegate();
        /// <summary>
        /// Refresh the log textbox in the form with our entire buffer. This has to runson the form's thread
        /// so it should be called only via Invoke
        /// </summary>
        private void UpdateWholeLog()
        {
            textBoxDebugLog.Clear();
            // using AppendText since that makes sure the TextBox stays
            // scrolled at the bottom
            textBoxDebugLog.AppendText(debugLog.ToString());
        }

        private void DebugLog_Load(object sender, System.EventArgs e)
        {
            // first time we create our form, so let's refresh our log window.
            this.Invoke(new UpdateWholeLogDelegate(UpdateWholeLog), null);
        }
    }
}
