using System;
using System.Collections.Generic;
using System.Text;
using SCLib;
using log4net;

namespace vpnsim.supplicant.simSupplicant
{
    /// <summary>
    /// Abstract class for representing a SIM Interface
    /// </summary>
    public abstract class SimInterface
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(SimInterface));
        protected ApduResponse response;

        /// <summary>
        /// Verify the CHV
        /// </summary>
        /// <param name="pinCode"></param>
        /// <returns></returns>
        public bool VerifyCHV(char[] pinCode)
        {
            response = TransmitAPDU(ApduCommands.VerifyCHV(1, pinCode));
            if (response.error) return false;
            return true;
        }

        /// <summary>
        /// Select a dedicated file
        /// </summary>
        /// <param name="file"></param>
        /// <returns></returns>
        public bool SelectFile(SIM.DedicatedFile file)
        {
            response = TransmitAPDU(ApduCommands.Select(file));
            if (response.error) return false;
            return true;
        }

        /// <summary>
        /// Select an elementary file
        /// </summary>
        /// <param name="file"></param>
        /// <returns></returns>
        public bool SelectFile(SIM.ElementaryFile file)
        {
            response = TransmitAPDU(ApduCommands.Select(file));
            if (response.error) return false;
            //response = TransmitAPDU(APDUCommand.GetResponse(response.SW2));
            return true;
        }

        /// <summary>
        /// Read binary data
        /// </summary>
        /// <param name="length"></param>
        /// <returns></returns>
        public bool ReadBinary(int length)
        {
            response = TransmitAPDU(ApduCommands.ReadBinary(0, 0, length));
            if (response.error) return false;
            return true;
        }

        /// <summary>
        /// Get the reponse we got from the sim
        /// </summary>
        /// <returns></returns>
        public ApduResponse GetResponse()
        {
            return response;
        }
        abstract public bool RunGSMAlgorithm(byte[] rand);
        abstract public ApduResponse TransmitAPDU(CardCommand command);
        abstract public void SetInterfaceName(string interfaceName);
        abstract public bool IsReady();
        abstract public bool Connect();
        abstract public bool Disconnect();
        abstract public bool Reset();
    }

    /// <summary>
    /// Class for keeping track of sim constants
    /// </summary>
    public class SIM
    {
        /// <summary>
        /// Enum for the Dedicatedfile addresses
        /// </summary>
        public enum DedicatedFile : short
        {
            TELECOM = 0x7F10,
            GSM = 0x7F20
        }

        /// <summary>
        /// Enum for the Elementary file addresses
        /// </summary>
        public enum ElementaryFile : short
        {
            ICCID = 0x2FE2,
            ADN = 0x6F3A,
            FDN = 0x6F3B,
            SMS = 0x6F3C,
            CCP = 0x6F3D,
            MSISDN = 0x6F40,
            SMSP = 0x6F42,
            SMSS = 0x6F43,
            LND = 0x6F44,
            EXT1 = 0x6F4A,
            EXT2 = 0x6F4B,
            LP = 0x6F05,
            IMSI = 0x6F07,
            Kc = 0x6F20,
            PLMNsel = 0x6F30,
            HPLMN = 0x6F31,
            ACMmax = 0x6F37,
            SST = 0x6F38,
            ACM = 0x6F39,
            GID1 = 0x6F3E,
            GID2 = 0x6F3F,
            PUCT = 0x6F41,
            CBMI = 0x6F45,
            SPN = 0x6F46,
            BCCH = 0x6F74,
            ACC = 0x6F78,
            FPLMN = 0x6F7B,
            LOCI = 0x6F7E,
            AD = 0x6FAD,
            PHASE = 0x6FAE
        }

        /// <summary>
        /// Enum for the record mode
        /// </summary>
        public enum recordMode
        {
            next = 0x02,
            previous = 0x03,
            absolute = 0x04
        }
    }
}
