using System;
using System.Collections.Generic;
using System.Net.Sockets;
using System.Text;
using System.Windows.Forms;
using InTheHand.Net;
using InTheHand.Net.Bluetooth;
using InTheHand.Net.Sockets;
using InTheHand.Windows.Forms;
using log4net;
using SCLib;

namespace vpnsim.supplicant.simSupplicant.bluetooth
{

    /// <summary>
    /// A class for communicating with a bluetooth SAP server
    /// </summary>
    public class BluetoothInterface : SimInterface
    {
        private static readonly ILog log = LogManager.GetLogger(typeof(BluetoothInterface));
        private BluetoothDeviceInfo btDevice;
        private BluetoothClient client;
        private BluetoothEndPoint bep;
        NetworkStream stream;
        public BluetoothInterface(){}

        /// <summary>
        /// Constructor for the BluetoothInterface
        /// </summary>
        /// <param name="deviceAddress">The address of the bluetooth device to connect to</param>
        public BluetoothInterface(string deviceAddress)
        {
            BluetoothAddress btAddr = BluetoothAddress.Parse(deviceAddress);
            this.btDevice = new BluetoothDeviceInfo(btAddr);
        }

        /// <summary>
        /// Show a dialog for selecting the bluetooth device
        /// </summary>
        /// <returns>The device address of the bluetooth device</returns>
        public static string SelectInterface()
        {
            SelectBluetoothDeviceDialog sbdd = new SelectBluetoothDeviceDialog();
            DialogResult res = sbdd.ShowDialog();
            return sbdd.SelectedDevice.DeviceAddress.ToString();
        }

        /// <summary>
        /// Set the address of the bluetooth device to connect to
        /// </summary>
        /// <param name="deviceAddress"></param>
        public override void SetInterfaceName(string deviceAddress)
        {
            BluetoothAddress btAddr = BluetoothAddress.Parse(deviceAddress);
            this.btDevice = new BluetoothDeviceInfo(btAddr);
        }
        
        /// <summary>
        /// Connect to the Bluetooth SAP Server
        /// </summary>
        /// <returns>If the connect was successful</returns>
        public override bool Connect()
        {

            bep = new BluetoothEndPoint(btDevice.DeviceAddress, BluetoothService.SimAccess);
            client = new BluetoothClient();
            //client.Authenticate = true;
            //client.Client.SetSocketOption(BluetoothSocketOptionLevel.RFComm, BluetoothSocketOptionName.XPAuthenticate, 1);
            try
            {
                client.Connect(bep);
                stream = client.GetStream();
            }catch(Exception)
            {
                log.Error("Bluetooth connection timed out, check that bluetooth is turned on");
                return false;
            }
            
            //object rawValue = client.Client.GetSocketOption(BluetoothSocketOptionLevel.RFComm, BluetoothSocketOptionName.GetAuthenticationEnabled);
            // = client.Client.GetSocketOption(BluetoothSocketOptionLevel.RFComm, BluetoothSocketOptionName.XPAuthenticate);
            //bool value = (bool)rawValue;  // ?or 'int'
            //log.Debug("socopt: " + value);
            
            if (!Connect_Request())
            {
                log.Info("Error connecting to SAP server");
                return false;
            }
            return true;
        }


        /// <summary>
        /// Get the device address of the interface
        /// </summary>
        /// <returns></returns>
        public string GetInterfaceName()
        {
            return btDevice.DeviceAddress.ToString();
        }
        private bool Connect_Request()
        {
            byte[] message = new byte[4 + 8];
            message[0] = 0x00; //Connect request msgid
            message[1] = 0x01; //Number of parameters (1, maxmsgsize
            message[2] = message[3] = 0x00; //Reserved
            message[4] = 0x00; //Maxmsgsize id
            message[5] = 0x00; //Reserved
            message[6] = 0x00; //Param length #1
            message[7] = 0x02; //Param length #2
            message[8] = 0x10; //maxmsgsize
            message[9] = 0x00;
            message[10] = 0x00; //Padding
            message[11] = 0x00; //Padding
            stream.Write(message, 0, message.Length);
            byte[] buffer = new byte[128];
            int bytesRead = stream.Read(buffer, 0, buffer.Length);
            Array.Resize(ref buffer, bytesRead);
            bool statusOK = false;
            if (buffer[0] == 0x01) //Connect_Resp
            { 
                if (buffer[8] == 0x00) statusOK = true; //This is connection request ok 
            }

            buffer = new byte[128];
            bytesRead = stream.Read(buffer, 0, buffer.Length);
            Array.Resize(ref buffer, bytesRead);

            return statusOK;

        }
        /// <summary>
        /// Send an APDU to the bluetooth sap server
        /// </summary>
        /// <param name="apdu"></param>
        /// <returns>The response from the server</returns>
        public override ApduResponse TransmitAPDU(CardCommand apdu)
        {
            
            byte[] apduBytes = apdu.GenerateBytes();
            //8 + apduBytes.Length + (4 - (8 + apduBytes.Length) % 4) % 4
            byte[] message = new byte[(((8+apduBytes.Length - 1) / 4) + 1) * 4];


            message[0] = 0x05; //APDU Req
            message[1] = 0x01; // num parameters
            message[4] = 0x04; //COmmand apdu id
            byte[] length = BitConverter.GetBytes((short)apduBytes.Length);
            message[6] = length[1];
            message[7] = length[0];
            Array.Copy(apduBytes, 0, message, 8, apduBytes.Length);
            stream.Write(message, 0, message.Length);
            byte[] buffer = new byte[128];
            int bytesRead = stream.Read(buffer, 0, buffer.Length);
            Array.Resize(ref buffer, bytesRead);
            if (buffer[0] != 0x06) //Did we receive an apdu response?
            {
                response = new ApduResponse(new byte[] {buffer[0], buffer[1] });
            }else{
                if (buffer[8] != 0x00) //IS response code != 0x00 (ok)?
                {
                    response = new ApduResponse(new byte[] { buffer[0], buffer[8] });
                }
                else
                { //Everything ok
                    length = new byte[2];
                    length[0] = buffer[15];
                    length[1] = buffer[14];
                    ushort len = BitConverter.ToUInt16(length,0);
                    byte[] apdures = new byte[len];
                    Array.Copy(buffer, 16, apdures, 0, len);
                    response = new ApduResponse(apdures);
                }
            }
            return response;
        }

        /// <summary>
        /// Run the GSM Algorithm via the SAP server, overridden, since the (my) bluetooth SAP server doesn't 
        /// require the get-response apdu, unlike other interfaces
        /// </summary>
        /// <param name="rand"></param>
        /// <returns></returns>
        public override bool RunGSMAlgorithm(byte[] rand)
        {
            response = TransmitAPDU(ApduCommands.RunGSMAlgorithm(rand));
            if (response.error) return false;
            //This one's weird on bluetooth
//            response = TransmitAPDU(ApduCommands.GetResponse(response.SW2));
            return true;
        }

        /// <summary>
        /// Disconnect from the Bluetooth SAP Server
        /// </summary>
        /// <returns></returns>
        public override bool Disconnect()
        {
            byte[] message = new byte[4];
            message[0] = 0x02; //DisconnectConnect request msgid
            stream.Write(message, 0, message.Length);
            stream.Close();
            return true;
        }

        /// <summary>
        /// Tells if the bluetooth interface is available
        /// </summary>
        /// <returns></returns>
        public static bool IsAvailable()
        {
            BluetoothRadio[] radios = BluetoothRadio.AllRadios;
            if (radios.Length == 0) return false;
            return true;
        }

        /// <summary>
        /// Checks to see if the interface is ready/existant
        /// </summary>
        /// <returns></returns>
        public override bool IsReady()
        {
            return BluetoothInterface.IsAvailable();
        }

        /// <summary>
        /// Reset the SIM, not implemented
        /// </summary>
        /// <returns></returns>
        public override bool Reset()
        {
            throw new NotImplementedException();
        }
    }
}
