using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using SCLib;
using vpnsim.supplicant;
using vpnsim.supplicant.simSupplicant;

namespace vpnsim.supplicant.simSupplicant.smartCard
{
    /// <summary>
    /// Class representing the Smart Card interface
    /// </summary>
    public class SmartCardInterface : SimInterface
    {
        private SCResMgr resMgr;
        private SCReader reader;
        private string readerName;

        #region Smart Card PCSC dll functions (windows)
        [DllImport("WinScard.dll")]
        public static extern int SCardEstablishContext(uint dwScope, int nNotUsed1, int nNotUsed2, ref int phContext);
        [DllImport("WinScard.dll")]
        public static extern int SCardReleaseContext(int phContext);
        [DllImport("WinScard.dll")]
        public static extern int SCardConnect(int hContext, string cReaderName, uint dwShareMode, uint dwPrefProtocol, ref int phCard, ref int ActiveProtocol);
        [DllImport("WinScard.dll")]
        public static extern int SCardDisconnect(int hCard, int Disposition);
        [DllImport("WinScard.dll")]
        public static extern int SCardListReaderGroups(int hContext, ref string cGroups, ref int nStringSize);
        [DllImport("WinScard.dll")]
        public static extern int SCardListReaders(int hContext, string cGroups, ref string cReaderLists, ref int nReaderCount);
        [DllImport("WinScard.dll")]
        public static extern int SCardFreeMemory(int hContext, string cResourceToFree);
        [DllImport("WinScard.dll")]
        public static extern int SCardGetAttrib(int hContext, uint dwAttrId, ref byte[] bytRecvAttr, ref int nRecLen);
        #endregion

        /// <summary>
        /// Constructor for the Smart Card Interface
        /// </summary>
        public SmartCardInterface()
        {
            resMgr = new SCResMgr();
            resMgr.EstablishContext(SCardContextScope.System);

        }

        /// <summary>
        /// Constructor overload for the smart card interface
        /// </summary>
        /// <param name="readerName">Name of the reader to use</param>
        public SmartCardInterface(string readerName) :this()
        {
            SetInterfaceName(readerName);
        }

        /// <summary>
        /// Shows a dialog letting the user select the smart card reader to use
        /// </summary>
        /// <returns>Name of the smart card reader</returns>
        public static string SelectInterface()
        {
            String selectedReader = SCReader.SelectReader();
            return selectedReader;
        }

        /// <summary>
        /// Returns the name of the current reader
        /// </summary>
        /// <returns></returns>
        public string GetInterfaceName()
        {
            return readerName;
        }

        /// <summary>
        /// Sets the name of the reader to use
        /// </summary>
        /// <param name="name"></param>
        public override void SetInterfaceName(string name)
        {
            this.readerName = name;
            reader = new SCReader(resMgr);
        }

        /// <summary>
        /// Connect to the SIM card in the reader
        /// </summary>
        /// <returns></returns>
        public override bool Connect()
        {
            bool result = reader.Connect(readerName, SCardAccessMode.Exclusive, SCardProtocolIdentifiers.T0);
            reader.BeginTransaction();
            return result;
            
        }

        /// <summary>
        /// Run the GSM Algorithm on the SIM in the reader
        /// </summary>
        /// <param name="rand">The RAND value</param>
        /// <returns>If the algorithm ran ok</returns>
        public override bool RunGSMAlgorithm(byte[] rand)
        {
            response = TransmitAPDU(ApduCommands.RunGSMAlgorithm(rand));
            if (response.error) return false;
            response = TransmitAPDU(ApduCommands.GetResponse(response.SW2));
            return true;
        }

        /// <summary>
        /// Disconnect from the SIM/Reader
        /// </summary>
        /// <returns></returns>
        public override bool Disconnect()
        {
            reader.EndTransaction(SCardDisposition.UnpowerCard);
            reader.Disconnect(SCardDisposition.UnpowerCard);
            return true;

        }

        /// <summary>
        /// Send an APDU to the SIM in the reader
        /// </summary>
        /// <param name="apdu">APDU to send</param>
        /// <returns>APDUResponse received from the SIM</returns>
        public override ApduResponse TransmitAPDU(CardCommand apdu)
        {
            byte[] res;
            reader.Transmit(apdu.GenerateBytes(), out res);
            response = new ApduResponse(res);
            return response;
        }


        /// <summary>
        /// Reset the SIM in the reader
        /// </summary>
        /// <returns></returns>
        public override bool Reset()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Check if a Smart Card reader is available
        /// </summary>
        /// <returns></returns>
        public static bool IsAvailable()
        {
            try
            {
                SCResMgr rm = new SCResMgr();
                rm.EstablishContext(SCardContextScope.System);
                rm.IsValidContext();
                List<string> test = new List<string>();
                if (rm.ListReaders(test) == 0) return false;
                
                
            }catch(Exception)
            {
                return false;
            }
            return true;
        }

        /// <summary>
        /// Is the interface ready?
        /// </summary>
        /// <returns></returns>
        public override bool IsReady()
        {
            return SmartCardInterface.IsAvailable();
        }
    }
}